# -*- coding: utf-8 -*-

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from google.cloud.aiplatform_v1beta1.services.dataset_service import (
    client as dataset_service_client_v1beta1,
)
from google.cloud.aiplatform_v1beta1.services.deployment_resource_pool_service import (
    client as deployment_resource_pool_service_client_v1beta1,
)
from google.cloud.aiplatform_v1beta1.services.endpoint_service import (
    client as endpoint_service_client_v1beta1,
)
from google.cloud.aiplatform_v1beta1.services.example_store_service import (
    client as example_store_service_client_v1beta1,
)
from google.cloud.aiplatform_v1beta1.services.extension_execution_service import (
    client as extension_execution_service_client_v1beta1,
)
from google.cloud.aiplatform_v1beta1.services.extension_registry_service import (
    client as extension_registry_service_client_v1beta1,
)
from google.cloud.aiplatform_v1beta1.services.feature_online_store_service import (
    client as feature_online_store_service_client_v1beta1,
)
from google.cloud.aiplatform_v1beta1.services.feature_online_store_admin_service import (
    client as feature_online_store_admin_service_client_v1beta1,
)
from google.cloud.aiplatform_v1beta1.services.feature_registry_service import (
    client as feature_registry_service_client_v1beta1,
)
from google.cloud.aiplatform_v1beta1.services.featurestore_online_serving_service import (
    client as featurestore_online_serving_service_client_v1beta1,
)
from google.cloud.aiplatform_v1beta1.services.featurestore_service import (
    client as featurestore_service_client_v1beta1,
)
from google.cloud.aiplatform_v1beta1.services.gen_ai_cache_service import (
    client as gen_ai_cache_service_client_v1beta1,
)
from google.cloud.aiplatform_v1beta1.services.index_service import (
    client as index_service_client_v1beta1,
)
from google.cloud.aiplatform_v1beta1.services.index_endpoint_service import (
    client as index_endpoint_service_client_v1beta1,
)
from google.cloud.aiplatform_v1beta1.services.job_service import (
    client as job_service_client_v1beta1,
)
from google.cloud.aiplatform_v1beta1.services.match_service import (
    client as match_service_client_v1beta1,
)
from google.cloud.aiplatform_v1beta1.services.metadata_service import (
    client as metadata_service_client_v1beta1,
)
from google.cloud.aiplatform_v1beta1.services.model_garden_service import (
    client as model_garden_service_client_v1beta1,
)
from google.cloud.aiplatform_v1beta1.services.model_service import (
    client as model_service_client_v1beta1,
)
from google.cloud.aiplatform_v1beta1.services.model_monitoring_service import (
    client as model_monitoring_service_client_v1beta1,
)
from google.cloud.aiplatform_v1beta1.services.persistent_resource_service import (
    client as persistent_resource_service_client_v1beta1,
)
from google.cloud.aiplatform_v1beta1.services.pipeline_service import (
    client as pipeline_service_client_v1beta1,
)
from google.cloud.aiplatform_v1beta1.services.prediction_service import (
    client as prediction_service_client_v1beta1,
)
from google.cloud.aiplatform_v1beta1.services.prediction_service import (
    async_client as prediction_service_async_client_v1beta1,
)
from google.cloud.aiplatform_v1beta1.services.reasoning_engine_service import (
    client as reasoning_engine_service_client_v1beta1,
)
from google.cloud.aiplatform_v1beta1.services.reasoning_engine_execution_service import (
    client as reasoning_engine_execution_service_client_v1beta1,
)
from google.cloud.aiplatform_v1beta1.services.schedule_service import (
    client as schedule_service_client_v1beta1,
)
from google.cloud.aiplatform_v1beta1.services.specialist_pool_service import (
    client as specialist_pool_service_client_v1beta1,
)
from google.cloud.aiplatform_v1beta1.services.tensorboard_service import (
    client as tensorboard_service_client_v1beta1,
)
from google.cloud.aiplatform_v1beta1.services.vertex_rag_data_service import (
    client as vertex_rag_data_service_client_v1beta1,
)
from google.cloud.aiplatform_v1beta1.services.vertex_rag_data_service import (
    async_client as vertex_rag_data_service_async_client_v1beta1,
)
from google.cloud.aiplatform_v1beta1.services.vertex_rag_service import (
    client as vertex_rag_service_client_v1beta1,
)
from google.cloud.aiplatform_v1beta1.services.vizier_service import (
    client as vizier_service_client_v1beta1,
)


from google.cloud.aiplatform_v1.services.dataset_service import (
    client as dataset_service_client_v1,
)
from google.cloud.aiplatform_v1.services.deployment_resource_pool_service import (
    client as deployment_resource_pool_service_client_v1,
)
from google.cloud.aiplatform_v1.services.endpoint_service import (
    client as endpoint_service_client_v1,
)
from google.cloud.aiplatform_v1.services.feature_online_store_service import (
    client as feature_online_store_service_client_v1,
)
from google.cloud.aiplatform_v1.services.feature_online_store_admin_service import (
    client as feature_online_store_admin_service_client_v1,
)
from google.cloud.aiplatform_v1.services.feature_registry_service import (
    client as feature_registry_service_client_v1,
)
from google.cloud.aiplatform_v1.services.featurestore_online_serving_service import (
    client as featurestore_online_serving_service_client_v1,
)
from google.cloud.aiplatform_v1.services.featurestore_service import (
    client as featurestore_service_client_v1,
)
from google.cloud.aiplatform_v1.services.gen_ai_cache_service import (
    client as gen_ai_cache_service_client_v1,
)
from google.cloud.aiplatform_v1.services.index_service import (
    client as index_service_client_v1,
)
from google.cloud.aiplatform_v1.services.index_endpoint_service import (
    client as index_endpoint_service_client_v1,
)
from google.cloud.aiplatform_v1.services.job_service import (
    client as job_service_client_v1,
)
from google.cloud.aiplatform_v1.services.metadata_service import (
    client as metadata_service_client_v1,
)
from google.cloud.aiplatform_v1.services.model_garden_service import (
    client as model_garden_service_client_v1,
)
from google.cloud.aiplatform_v1.services.model_service import (
    client as model_service_client_v1,
)
from google.cloud.aiplatform_v1.services.persistent_resource_service import (
    client as persistent_resource_service_client_v1,
)
from google.cloud.aiplatform_v1.services.pipeline_service import (
    client as pipeline_service_client_v1,
)
from google.cloud.aiplatform_v1.services.prediction_service import (
    client as prediction_service_client_v1,
)
from google.cloud.aiplatform_v1.services.prediction_service import (
    async_client as prediction_service_async_client_v1,
)
from google.cloud.aiplatform_v1.services.reasoning_engine_service import (
    client as reasoning_engine_service_client_v1,
)
from google.cloud.aiplatform_v1.services.reasoning_engine_execution_service import (
    client as reasoning_engine_execution_service_client_v1,
)
from google.cloud.aiplatform_v1.services.reasoning_engine_execution_service import (
    async_client as reasoning_engine_execution_async_client_v1,
)
from google.cloud.aiplatform_v1.services.schedule_service import (
    client as schedule_service_client_v1,
)
from google.cloud.aiplatform_v1.services.specialist_pool_service import (
    client as specialist_pool_service_client_v1,
)
from google.cloud.aiplatform_v1.services.tensorboard_service import (
    client as tensorboard_service_client_v1,
)
from google.cloud.aiplatform_v1.services.vizier_service import (
    client as vizier_service_client_v1,
)
from google.cloud.aiplatform_v1.services.vertex_rag_service import (
    client as vertex_rag_service_client_v1,
)
from google.cloud.aiplatform_v1.services.vertex_rag_data_service import (
    client as vertex_rag_data_service_client_v1,
)
from google.cloud.aiplatform_v1.services.vertex_rag_data_service import (
    async_client as vertex_rag_data_service_async_client_v1,
)

__all__ = (
    # v1
    dataset_service_client_v1,
    deployment_resource_pool_service_client_v1,
    endpoint_service_client_v1,
    feature_online_store_service_client_v1,
    feature_online_store_admin_service_client_v1,
    feature_registry_service_client_v1,
    featurestore_online_serving_service_client_v1,
    featurestore_service_client_v1,
    index_service_client_v1,
    index_endpoint_service_client_v1,
    job_service_client_v1,
    metadata_service_client_v1,
    model_garden_service_client_v1,
    model_service_client_v1,
    persistent_resource_service_client_v1,
    pipeline_service_client_v1,
    prediction_service_client_v1,
    prediction_service_async_client_v1,
    reasoning_engine_execution_service_client_v1,
    reasoning_engine_service_client_v1,
    schedule_service_client_v1,
    specialist_pool_service_client_v1,
    tensorboard_service_client_v1,
    vizier_service_client_v1,
    vertex_rag_data_service_async_client_v1,
    vertex_rag_data_service_client_v1,
    vertex_rag_service_client_v1,
    # v1beta1
    dataset_service_client_v1beta1,
    deployment_resource_pool_service_client_v1beta1,
    endpoint_service_client_v1beta1,
    example_store_service_client_v1beta1,
    feature_online_store_service_client_v1beta1,
    feature_online_store_admin_service_client_v1beta1,
    feature_registry_service_client_v1beta1,
    featurestore_online_serving_service_client_v1beta1,
    featurestore_service_client_v1beta1,
    index_service_client_v1beta1,
    index_endpoint_service_client_v1beta1,
    job_service_client_v1beta1,
    match_service_client_v1beta1,
    model_garden_service_client_v1beta1,
    model_monitoring_service_client_v1beta1,
    model_service_client_v1beta1,
    persistent_resource_service_client_v1beta1,
    pipeline_service_client_v1beta1,
    prediction_service_client_v1beta1,
    prediction_service_async_client_v1beta1,
    reasoning_engine_execution_service_client_v1beta1,
    reasoning_engine_service_client_v1beta1,
    schedule_service_client_v1beta1,
    specialist_pool_service_client_v1beta1,
    metadata_service_client_v1beta1,
    tensorboard_service_client_v1beta1,
    vertex_rag_service_client_v1beta1,
    vertex_rag_data_service_client_v1beta1,
    vertex_rag_data_service_async_client_v1beta1,
    vizier_service_client_v1beta1,
)



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Utility functions for prompt optimizer."""

import json
from . import types


def _get_service_account(
    config: types.PromptOptimizerConfigOrDict,
) -> str:
    """Get the service account from the config for the custom job."""
    if isinstance(config, dict):
        config = types.PromptOptimizerConfig.model_validate(config)

    if config.service_account and config.service_account_project_number:
        raise ValueError(
            "Only one of service_account or "
            "service_account_project_number can be provided."
        )
    elif config.service_account:
        return config.service_account
    elif config.service_account_project_number:
        return f"{config.service_account_project_number}-compute@developer.gserviceaccount.com"
    else:
        raise ValueError(
            "Either service_account or service_account_project_number " "is required."
        )


def _clean_and_parse_optimized_prompt(output_str: str):
    """Cleans a string response returned from the prompt optimizer endpoint.

    Args:
        output_str: The optimized prompt string containing the JSON data,
          potentially with markdown formatting like ```json ... ```.

    Returns:
        The parsed JSON data, or None if parsing fails.
    """
    lines = output_str.strip().split("\n")
    # Remove markdown delimiters
    if lines and lines[0].strip().startswith("```"):
        cleaned_string = "\n".join(lines[1:-1])
    else:
        cleaned_string = output_str

    # remove any 'json' labels if they exist on the first line.
    if cleaned_string.strip().startswith("json"):
        cleaned_string = cleaned_string.strip()[4:].strip()

    try:
        return json.loads(cleaned_string)
    except json.JSONDecodeError as e:
        # TODO(b/437144880): raise errors.ClientError here instead
        raise ValueError(
            f"Failed to parse the response from prompt optimizer endpoint. {e}"
        ) from e


def _parse(output_str: str) -> types.ParsedResponse:
    """Parses the output string from the prompt optimizer endpoint."""
    parsed_out = _clean_and_parse_optimized_prompt(output_str)
    return types.ParsedResponse(**parsed_out)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""User friendly container for Cloud Spanner Table."""

from google.cloud.exceptions import NotFound

from google.cloud.spanner_admin_database_v1 import DatabaseDialect
from google.cloud.spanner_v1.types import (
    Type,
    TypeCode,
)


_EXISTS_TEMPLATE = """
SELECT EXISTS(
    SELECT TABLE_NAME
    FROM INFORMATION_SCHEMA.TABLES
    {}
)
"""
_GET_SCHEMA_TEMPLATE = "SELECT * FROM {} LIMIT 0"


class Table(object):
    """Representation of a Cloud Spanner Table.

    :type table_id: str
    :param table_id: The ID of the table.

    :type database: :class:`~google.cloud.spanner_v1.database.Database`
    :param database: The database that owns the table.
    """

    def __init__(self, table_id, database, schema_name=None):
        if schema_name is None:
            self._schema_name = database.default_schema_name
        else:
            self._schema_name = schema_name
        self._table_id = table_id
        self._database = database

        # Calculated properties.
        self._schema = None

    @property
    def schema_name(self):
        """The schema name of the table used in SQL.

        :rtype: str
        :returns: The table schema name.
        """
        return self._schema_name

    @property
    def table_id(self):
        """The ID of the table used in SQL.

        :rtype: str
        :returns: The table ID.
        """
        return self._table_id

    @property
    def qualified_table_name(self):
        """The qualified name of the table used in SQL.

        :rtype: str
        :returns: The qualified table name.
        """
        if self.schema_name == self._database.default_schema_name:
            return self._quote_identifier(self.table_id)
        return "{}.{}".format(
            self._quote_identifier(self.schema_name),
            self._quote_identifier(self.table_id),
        )

    def _quote_identifier(self, identifier):
        """Quotes the given identifier using the rules of the dialect of the database of this table.

        :rtype: str
        :returns: The quoted identifier.
        """
        if self._database.database_dialect == DatabaseDialect.POSTGRESQL:
            return '"{}"'.format(identifier)
        return "`{}`".format(identifier)

    def exists(self):
        """Test whether this table exists.

        :rtype: bool
        :returns: True if the table exists, else false.
        """
        with self._database.snapshot() as snapshot:
            return self._exists(snapshot)

    def _exists(self, snapshot):
        """Query to check that the table exists.

        :type snapshot: :class:`~google.cloud.spanner_v1.snapshot.Snapshot`
        :param snapshot: snapshot to use for database queries

        :rtype: bool
        :returns: True if the table exists, else false.
        """
        if self._database.database_dialect == DatabaseDialect.POSTGRESQL:
            results = snapshot.execute_sql(
                sql=_EXISTS_TEMPLATE.format(
                    "WHERE TABLE_SCHEMA=$1 AND TABLE_NAME = $2"
                ),
                params={"p1": self.schema_name, "p2": self.table_id},
                param_types={
                    "p1": Type(code=TypeCode.STRING),
                    "p2": Type(code=TypeCode.STRING),
                },
            )
        else:
            results = snapshot.execute_sql(
                sql=_EXISTS_TEMPLATE.format(
                    "WHERE TABLE_SCHEMA = @schema_name AND TABLE_NAME = @table_id"
                ),
                params={"schema_name": self.schema_name, "table_id": self.table_id},
                param_types={
                    "schema_name": Type(code=TypeCode.STRING),
                    "table_id": Type(code=TypeCode.STRING),
                },
            )
        return next(iter(results))[0]

    @property
    def schema(self):
        """The schema of this table.

        :rtype: list of :class:`~google.cloud.spanner_v1.types.StructType.Field`
        :returns: The table schema.
        """
        if self._schema is None:
            with self._database.snapshot() as snapshot:
                self._schema = self._get_schema(snapshot)
        return self._schema

    def _get_schema(self, snapshot):
        """Get the schema of this table.

        :type snapshot: :class:`~google.cloud.spanner_v1.snapshot.Snapshot`
        :param snapshot: snapshot to use for database queries

        :rtype: list of :class:`~google.cloud.spanner_v1.types.StructType.Field`
        :returns: The table schema.
        """
        query = _GET_SCHEMA_TEMPLATE.format(self.qualified_table_name)
        results = snapshot.execute_sql(query)
        # Start iterating to force the schema to download.
        try:
            next(iter(results))
        except StopIteration:
            pass
        return list(results.fields)

    def reload(self):
        """Reload this table.

        Refresh any configured schema into :attr:`schema`.

        :raises NotFound: if the table does not exist
        """
        with self._database.snapshot() as snapshot:
            if not self._exists(snapshot):
                raise NotFound("table '{}' does not exist".format(self.table_id))
            self._schema = self._get_schema(snapshot)



#!/usr/bin/env python3
"""
BMasterAI Qdrant Cloud RAG with Gradio UI
Interactive web interface for RAG system using Qdrant Cloud and BMasterAI framework
"""

import os
import json
import time
import uuid
import hashlib
from typing import List, Dict, Any, Optional, Tuple
from dataclasses import dataclass
from datetime import datetime
import threading

# BMasterAI imports
from bmasterai.logging import configure_logging, get_logger, LogLevel, EventType
from bmasterai.monitoring import get_monitor

# External dependencies
try:
    import gradio as gr
    import openai
    from qdrant_client import QdrantClient
    from qdrant_client.models import Distance, VectorParams, PointStruct, Filter, FieldCondition, MatchValue
    import numpy as np
    from sentence_transformers import SentenceTransformer
    import pandas as pd
except ImportError as e:
    print(f"Missing required dependencies: {e}")
    print("Install with: pip install gradio openai qdrant-client sentence-transformers numpy pandas")
    exit(1)

@dataclass
class QdrantConfig:
    """Configuration for Qdrant Cloud connection"""
    url: str
    api_key: str
    collection_name: str = "bmasterai_knowledge"
    vector_size: int = 384  # all-MiniLM-L6-v2 embedding size
    distance: Distance = Distance.COSINE
    timeout: int = 30

@dataclass
class RAGConfig:
    """Configuration for RAG system"""
    openai_api_key: str
    embedding_model: str = "all-MiniLM-L6-v2"
    llm_model: str = "gpt-3.5-turbo"
    max_tokens: int = 1000
    temperature: float = 0.7
    top_k_results: int = 5
    similarity_threshold: float = 0.7

class BMasterAIQdrantRAGUI:
    """
    BMasterAI Qdrant RAG system with Gradio UI
    """
    
    def __init__(self, qdrant_config: QdrantConfig, rag_config: RAGConfig):
        self.qdrant_config = qdrant_config
        self.rag_config = rag_config
        
        # Initialize BMasterAI components
        self.logger = get_logger()
        self.monitor = get_monitor()
        self.agent_id = "qdrant-rag-ui"
        
        # Initialize clients
        self.qdrant_client = None
        self.openai_client = None
        self.embedding_model = None
        
        # UI state
        self.chat_history = []
        self.system_status = "Initializing..."
        self.collection_info = {}
        
        # Initialize system
        self._initialize_system()
    
    def _initialize_system(self):
        """Initialize all system components"""
        try:
            self.system_status = "🔄 Initializing Qdrant client..."
            self._init_qdrant_client()
            
            self.system_status = "🔄 Initializing OpenAI client..."
            self._init_openai_client()
            
            self.system_status = "🔄 Loading embedding model..."
            self._init_embedding_model()
            
            self.system_status = "🔄 Setting up collection..."
            self.create_collection()
            
            self.system_status = "✅ System ready!"
            
            self.logger.log_event(
                self.agent_id,
                EventType.AGENT_START,
                "BMasterAI Qdrant RAG UI system initialized successfully",
                metadata={
                    "qdrant_url": self.qdrant_config.url,
                    "collection": self.qdrant_config.collection_name,
                    "embedding_model": self.rag_config.embedding_model,
                    "llm_model": self.rag_config.llm_model
                }
            )
            
        except Exception as e:
            self.system_status = f"❌ Initialization failed: {str(e)}"
            self.logger.log_event(
                self.agent_id,
                EventType.TASK_ERROR,
                f"System initialization failed: {str(e)}",
                level=LogLevel.ERROR
            )
    
    def _init_qdrant_client(self):
        """Initialize Qdrant Cloud client"""
        self.qdrant_client = QdrantClient(
            url=self.qdrant_config.url,
            api_key=self.qdrant_config.api_key,
            timeout=self.qdrant_config.timeout
        )
        
        # Test connection
        collections = self.qdrant_client.get_collections()
        self.logger.log_event(
            self.agent_id,
            EventType.AGENT_COMMUNICATION,
            "Connected to Qdrant Cloud successfully",
            metadata={"collections_count": len(collections.collections)}
        )
    
    def _init_openai_client(self):
        """Initialize OpenAI client"""
        openai.api_key = self.rag_config.openai_api_key
        self.openai_client = openai
        
        self.logger.log_event(
            self.agent_id,
            EventType.AGENT_COMMUNICATION,
            "OpenAI client initialized",
            metadata={"model": self.rag_config.llm_model}
        )
    
    def _init_embedding_model(self):
        """Initialize sentence transformer model for embeddings"""
        self.embedding_model = SentenceTransformer(self.rag_config.embedding_model)
        
        self.logger.log_event(
            self.agent_id,
            EventType.AGENT_START,
            "Embedding model loaded successfully",
            metadata={"model": self.rag_config.embedding_model}
        )
    
    def create_collection(self) -> str:
        """Create Qdrant collection if it doesn't exist"""
        try:
            # Check if collection exists
            collections = self.qdrant_client.get_collections()
            collection_names = [col.name for col in collections.collections]
            
            if self.qdrant_config.collection_name in collection_names:
                message = f"✅ Collection '{self.qdrant_config.collection_name}' already exists"
                self.logger.log_event(
                    self.agent_id,
                    EventType.TASK_COMPLETE,
                    message
                )
                self._update_collection_info()
                return message
            
            # Create collection
            self.qdrant_client.create_collection(
                collection_name=self.qdrant_config.collection_name,
                vectors_config=VectorParams(
                    size=self.qdrant_config.vector_size,
                    distance=self.qdrant_config.distance
                )
            )
            
            message = f"✅ Created collection '{self.qdrant_config.collection_name}'"
            self.logger.log_event(
                self.agent_id,
                EventType.TASK_COMPLETE,
                message,
                metadata={
                    "vector_size": self.qdrant_config.vector_size,
                    "distance": self.qdrant_config.distance.value
                }
            )
            
            self._update_collection_info()
            return message
            
        except Exception as e:
            error_msg = f"❌ Failed to create collection: {str(e)}"
            self.logger.log_event(
                self.agent_id,
                EventType.TASK_ERROR,
                error_msg,
                level=LogLevel.ERROR
            )
            return error_msg
    
    def _update_collection_info(self):
        """Update collection information"""
        try:
            collection_info = self.qdrant_client.get_collection(self.qdrant_config.collection_name)
            self.collection_info = {
                "name": self.qdrant_config.collection_name,
                "vector_size": collection_info.config.params.vectors.size,
                "distance": collection_info.config.params.vectors.distance.value,
                "points_count": collection_info.points_count,
                "status": collection_info.status.value
            }
        except Exception as e:
            self.collection_info = {"error": str(e)}
    
    def add_documents_from_text(self, documents_text: str, source_name: str = "manual_input") -> str:
        """
        Add documents from text input (one document per line)
        """
        try:
            if not documents_text.strip():
                return "❌ Please provide document text"
            
            # Split text into documents (by lines or paragraphs)
            lines = [line.strip() for line in documents_text.split('\n') if line.strip()]
            
            if not lines:
                return "❌ No valid documents found"
            
            documents = []
            for i, line in enumerate(lines):
                documents.append({
                    "text": line,
                    "metadata": {
                        "source": source_name,
                        "line_number": i + 1,
                        "added_at": datetime.now().isoformat()
                    },
                    "source": source_name
                })
            
            result = self.add_documents(documents)
            self._update_collection_info()
            return result
            
        except Exception as e:
            error_msg = f"❌ Error processing documents: {str(e)}"
            self.logger.log_event(
                self.agent_id,
                EventType.TASK_ERROR,
                error_msg,
                level=LogLevel.ERROR
            )
            return error_msg
    
    def add_documents(self, documents: List[Dict[str, Any]]) -> str:
        """Add documents to Qdrant collection"""
        start_time = time.time()
        task_id = str(uuid.uuid4())
        
        try:
            self.logger.log_event(
                self.agent_id,
                EventType.TASK_START,
                f"Adding {len(documents)} documents to Qdrant",
                metadata={"task_id": task_id, "document_count": len(documents)}
            )
            
            points = []
            
            for i, doc in enumerate(documents):
                # Generate embedding
                embedding = self.embedding_model.encode(doc['text']).tolist()
                
                # Create unique ID
                doc_id = hashlib.md5(doc['text'].encode()).hexdigest()
                
                # Create point
                point = PointStruct(
                    id=doc_id,
                    vector=embedding,
                    payload={
                        "text": doc['text'],
                        "metadata": doc.get('metadata', {}),
                        "timestamp": datetime.now().isoformat(),
                        "source": doc.get('source', 'unknown')
                    }
                )
                points.append(point)
            
            # Upload to Qdrant
            self.qdrant_client.upsert(
                collection_name=self.qdrant_config.collection_name,
                points=points
            )
            
            duration_ms = (time.time() - start_time) * 1000
            
            success_msg = f"✅ Successfully added {len(documents)} documents in {duration_ms:.0f}ms"
            
            self.logger.log_event(
                self.agent_id,
                EventType.TASK_COMPLETE,
                success_msg,
                duration_ms=duration_ms,
                metadata={
                    "task_id": task_id,
                    "document_count": len(documents),
                    "collection": self.qdrant_config.collection_name
                }
            )
            
            self.monitor.track_task_duration(self.agent_id, "add_documents", duration_ms)
            return success_msg
            
        except Exception as e:
            duration_ms = (time.time() - start_time) * 1000
            error_msg = f"❌ Failed to add documents: {str(e)}"
            
            self.logger.log_event(
                self.agent_id,
                EventType.TASK_ERROR,
                error_msg,
                level=LogLevel.ERROR,
                duration_ms=duration_ms,
                metadata={"task_id": task_id}
            )
            
            self.monitor.track_error(self.agent_id, "add_documents")
            return error_msg
    
    def search_similar(self, query: str, limit: int = None) -> List[Dict[str, Any]]:
        """Search for similar documents in Qdrant"""
        start_time = time.time()
        
        if limit is None:
            limit = self.rag_config.top_k_results
        
        try:
            # Generate query embedding
            query_embedding = self.embedding_model.encode(query).tolist()
            
            # Search in Qdrant
            search_results = self.qdrant_client.search(
                collection_name=self.qdrant_config.collection_name,
                query_vector=query_embedding,
                limit=limit,
                score_threshold=self.rag_config.similarity_threshold
            )
            
            # Format results
            results = []
            for result in search_results:
                results.append({
                    "id": result.id,
                    "score": result.score,
                    "text": result.payload["text"],
                    "metadata": result.payload.get("metadata", {}),
                    "source": result.payload.get("source", "unknown"),
                    "timestamp": result.payload.get("timestamp")
                })
            
            duration_ms = (time.time() - start_time) * 1000
            
            self.logger.log_event(
                self.agent_id,
                EventType.TASK_COMPLETE,
                f"Found {len(results)} similar documents",
                duration_ms=duration_ms,
                metadata={
                    "query_length": len(query),
                    "results_count": len(results),
                    "avg_score": np.mean([r["score"] for r in results]) if results else 0
                }
            )
            
            self.monitor.track_task_duration(self.agent_id, "search_similar", duration_ms)
            return results
            
        except Exception as e:
            self.logger.log_event(
                self.agent_id,
                EventType.TASK_ERROR,
                f"Failed to search documents: {str(e)}",
                level=LogLevel.ERROR
            )
            
            self.monitor.track_error(self.agent_id, "search_similar")
            return []
    
    def generate_answer(self, query: str, context_docs: List[Dict[str, Any]]) -> str:
        """Generate answer using OpenAI with retrieved context"""
        start_time = time.time()
        
        try:
            # Prepare context
            context_text = "\n\n".join([
                f"Document {i+1} (Score: {doc['score']:.3f}):\n{doc['text']}"
                for i, doc in enumerate(context_docs)
            ])
            
            # Create prompt
            prompt = f"""Based on the following context documents, please answer the question. If the answer cannot be found in the context, please say so.

Context:
{context_text}

Question: {query}

Answer:"""
            
            # Call OpenAI
            response = self.openai_client.chat.completions.create(
                model=self.rag_config.llm_model,
                messages=[
                    {"role": "system", "content": "You are a helpful assistant that answers questions based on provided context."},
                    {"role": "user", "content": prompt}
                ],
                max_tokens=self.rag_config.max_tokens,
                temperature=self.rag_config.temperature
            )
            
            answer = response.choices[0].message.content
            tokens_used = response.usage.total_tokens
            
            duration_ms = (time.time() - start_time) * 1000
            
            self.logger.log_event(
                self.agent_id,
                EventType.LLM_CALL,
                "Answer generated successfully",
                duration_ms=duration_ms,
                metadata={
                    "tokens_used": tokens_used,
                    "answer_length": len(answer),
                    "context_length": len(context_text)
                }
            )
            
            self.monitor.track_llm_call(
                self.agent_id,
                self.rag_config.llm_model,
                tokens_used,
                duration_ms
            )
            
            return answer
            
        except Exception as e:
            error_msg = f"Error generating answer: {str(e)}"
            
            self.logger.log_event(
                self.agent_id,
                EventType.TASK_ERROR,
                error_msg,
                level=LogLevel.ERROR
            )
            
            self.monitor.track_error(self.agent_id, "generate_answer")
            return error_msg
    
    def chat_interface(self, message: str, history: List[List[str]]) -> Tuple[str, List[List[str]]]:
        """Main chat interface for Gradio"""
        if not message.strip():
            return "", history
        
        try:
            # Step 1: Search for similar documents
            similar_docs = self.search_similar(message)
            
            if not similar_docs:
                response = "I couldn't find relevant information to answer your question. Please try adding some documents first."
                history.append([message, response])
                return "", history
            
            # Step 2: Generate answer
            answer = self.generate_answer(message, similar_docs)
            
            # Add sources information
            sources_info = f"\n\n📚 **Sources ({len(similar_docs)} documents found):**\n"
            for i, doc in enumerate(similar_docs[:3]):  # Show top 3 sources
                sources_info += f"• **Source {i+1}** (Score: {doc['score']:.3f}): {doc['text'][:100]}...\n"
            
            full_response = answer + sources_info
            history.append([message, full_response])
            
            return "", history
            
        except Exception as e:
            error_response = f"Error processing your question: {str(e)}"
            history.append([message, error_response])
            return "", history
    
    def get_system_status(self) -> str:
        """Get current system status"""
        try:
            self._update_collection_info()
            
            status_info = f"""
## 🖥️ System Status: {self.system_status}

### 📊 Collection Information:
- **Name**: {self.collection_info.get('name', 'N/A')}
- **Documents**: {self.collection_info.get('points_count', 0)}
- **Vector Size**: {self.collection_info.get('vector_size', 'N/A')}
- **Distance Metric**: {self.collection_info.get('distance', 'N/A')}
- **Status**: {self.collection_info.get('status', 'N/A')}

### ⚙️ Configuration:
- **Embedding Model**: {self.rag_config.embedding_model}
- **LLM Model**: {self.rag_config.llm_model}
- **Top-K Results**: {self.rag_config.top_k_results}
- **Similarity Threshold**: {self.rag_config.similarity_threshold}

### 🔗 Connections:
- **Qdrant Cloud**: {'✅ Connected' if self.qdrant_client else '❌ Not Connected'}
- **OpenAI API**: {'✅ Connected' if self.openai_client else '❌ Not Connected'}
- **Embedding Model**: {'✅ Loaded' if self.embedding_model else '❌ Not Loaded'}
"""
            return status_info
            
        except Exception as e:
            return f"❌ Error getting system status: {str(e)}"
    
    def get_performance_metrics(self) -> str:
        """Get performance metrics from BMasterAI monitoring"""
        try:
            dashboard = self.monitor.get_agent_dashboard(self.agent_id)
            
            metrics_info = f"""
## 📈 Performance Metrics

### 🎯 Task Performance:
"""
            
            if dashboard.get('performance'):
                for task_name, metrics in dashboard['performance'].items():
                    metrics_info += f"""
**{task_name.replace('_', ' ').title()}:**
- Average Duration: {metrics.get('avg_duration_ms', 0):.2f}ms
- Min Duration: {metrics.get('min_duration_ms', 0):.2f}ms
- Max Duration: {metrics.get('max_duration_ms', 0):.2f}ms
- Total Calls: {metrics.get('total_calls', 0)}
"""
            
            metrics_info += f"""
### 📊 System Metrics:
- **Total Events**: {dashboard.get('metrics', {}).get('total_errors', 0)}
- **Error Count**: {dashboard.get('metrics', {}).get('total_errors', 0)}
- **Agent Status**: {dashboard.get('status', 'Unknown')}
"""
            
            return metrics_info
            
        except Exception as e:
            return f"❌ Error getting performance metrics: {str(e)}"
    
    def create_gradio_interface(self) -> gr.Blocks:
        """Create the Gradio interface"""
        
        # Custom CSS
        custom_css = """
        .gradio-container {
            max-width: 1400px;
            margin: auto;
        }
        .header-text {
            text-align: center;
            font-size: 2.5em;
            font-weight: bold;
            margin-bottom: 20px;
            background: linear-gradient(45deg, #FF6B6B, #4ECDC4);
            -webkit-background-clip: text;
            -webkit-text-fill-color: transparent;
        }
        .status-box {
            background-color: #f8f9fa;
            padding: 15px;
            border-radius: 10px;
            border-left: 4px solid #007bff;
            margin: 10px 0;
        }
        .metric-box {
            background-color: #e8f5e8;
            padding: 15px;
            border-radius: 10px;
            border-left: 4px solid #28a745;
            margin: 10px 0;
        }
        """
        
        with gr.Blocks(css=custom_css, title="BMasterAI Qdrant RAG", theme=gr.themes.Soft()) as demo:
            
            # Header
            gr.HTML("""
            <div class="header-text">
                🚀 BMasterAI Qdrant Cloud RAG System
            </div>
            <div style="text-align: center; margin-bottom: 30px;">
                <p style="font-size: 1.2em; color: #666;">
                    Advanced Retrieval-Augmented Generation with Qdrant Cloud & BMasterAI Framework
                </p>
            </div>
            """)
            
            with gr.Tabs():
                
                # Chat Tab
                with gr.TabItem("💬 Chat Interface", elem_id="chat-tab"):
                    with gr.Row():
                        with gr.Column(scale=3):
                            chatbot = gr.Chatbot(
                                label="RAG Chat Assistant",
                                height=500,
                                show_label=True,
                                elem_id="chatbot",
                                bubble_full_width=False
                            )
                            
                            with gr.Row():
                                msg = gr.Textbox(
                                    label="Your Question",
                                    placeholder="Ask me anything about your documents...",
                                    lines=2,
                                    scale=4
                                )
                                submit_btn = gr.Button("Send", variant="primary", scale=1)
                            
                            with gr.Row():
                                clear_btn = gr.Button("Clear Chat", variant="secondary")
                                example_btn = gr.Button("Load Example Questions", variant="secondary")
                        
                        with gr.Column(scale=1):
                            gr.HTML("<h3>💡 Quick Actions</h3>")
                            
                            # Example questions
                            with gr.Accordion("Example Questions", open=True):
                                example_questions = [
                                    "What is BMasterAI?",
                                    "How does RAG work?",
                                    "What are vector embeddings?",
                                    "Tell me about Qdrant database"
                                ]
                                
                                for question in example_questions:
                                    example_q_btn = gr.Button(f"❓ {question}", size="sm")
                                    example_q_btn.click(
                                        lambda q=question: (q, []),
                                        outputs=[msg, chatbot]
                                    )
                            
                            # Quick stats
                            with gr.Accordion("Quick Stats", open=False):
                                stats_display = gr.Markdown("Loading stats...")
                                refresh_stats_btn = gr.Button("Refresh Stats", size="sm")
                
                # Document Management Tab
                with gr.TabItem("📚 Document Management", elem_id="docs-tab"):
                    with gr.Row():
                        with gr.Column():
                            gr.HTML("<h3>📝 Add Documents</h3>")
                            
                            doc_input = gr.Textbox(
                                label="Documents (one per line)",
                                placeholder="Enter your documents here, one per line...\n\nExample:\nBMasterAI is an advanced AI framework.\nQdrant is a vector database for similarity search.\nRAG combines retrieval with generation.",
                                lines=10
                            )
                            
                            source_name = gr.Textbox(
                                label="Source Name",
                                placeholder="e.g., 'knowledge_base', 'manual_input', 'documentation'",
                                value="manual_input"
                            )
                            
                            add_docs_btn = gr.Button("Add Documents", variant="primary")
                            add_result = gr.Textbox(label="Result", interactive=False)
                        
                        with gr.Column():
                            gr.HTML("<h3>📊 Collection Status</h3>")
                            
                            collection_status = gr.Markdown("Loading collection status...")
                            refresh_collection_btn = gr.Button("Refresh Collection Info")
                            
                            gr.HTML("<h3>🔍 Search Test</h3>")
                            
                            search_query = gr.Textbox(
                                label="Test Search Query",
                                placeholder="Enter a query to test document retrieval..."
                            )
                            search_btn = gr.Button("Search Documents")
                            search_results = gr.JSON(label="Search Results")
                
                # System Status Tab
                with gr.TabItem("🖥️ System Status", elem_id="status-tab"):
                    with gr.Row():
                        with gr.Column():
                            system_status_display = gr.Markdown("Loading system status...")
                            refresh_status_btn = gr.Button("Refresh Status", variant="primary")
                        
                        with gr.Column():
                            performance_metrics = gr.Markdown("Loading performance metrics...")
                            refresh_metrics_btn = gr.Button("Refresh Metrics", variant="primary")
                
                # Configuration Tab
                with gr.TabItem("⚙️ Configuration", elem_id="config-tab"):
                    with gr.Row():
                        with gr.Column():
                            gr.HTML("<h3>🔧 RAG Settings</h3>")
                            
                            top_k_slider = gr.Slider(
                                minimum=1,
                                maximum=20,
                                value=self.rag_config.top_k_results,
                                step=1,
                                label="Top-K Results"
                            )
                            
                            similarity_threshold_slider = gr.Slider(
                                minimum=0.0,
                                maximum=1.0,
                                value=self.rag_config.similarity_threshold,
                                step=0.05,
                                label="Similarity Threshold"
                            )
                            
                            temperature_slider = gr.Slider(
                                minimum=0.0,
                                maximum=1.0,
                                value=self.rag_config.temperature,
                                step=0.1,
                                label="LLM Temperature"
                            )
                            
                            max_tokens_slider = gr.Slider(
                                minimum=100,
                                maximum=2000,
                                value=self.rag_config.max_tokens,
                                step=100,
                                label="Max Tokens"
                            )
                            
                            update_config_btn = gr.Button("Update Configuration", variant="primary")
                            config_result = gr.Textbox(label="Configuration Update Result", interactive=False)
                        
                        with gr.Column():
                            gr.HTML("<h3>📋 Current Configuration</h3>")
                            
                            current_config = gr.JSON(
                                value={
                                    "qdrant_url": self.qdrant_config.url,
                                    "collection_name": self.qdrant_config.collection_name,
                                    "embedding_model": self.rag_config.embedding_model,
                                    "llm_model": self.rag_config.llm_model,
                                    "top_k_results": self.rag_config.top_k_results,
                                    "similarity_threshold": self.rag_config.similarity_threshold,
                                    "temperature": self.rag_config.temperature,
                                    "max_tokens": self.rag_config.max_tokens
                                },
                                label="Current Settings"
                            )
            
            # Event handlers
            def submit_message(message, history):
                if message.strip():
                    return self.chat_interface(message, history)
                return message, history
            
            def update_configuration(top_k, similarity_threshold, temperature, max_tokens):
                try:
                    self.rag_config.top_k_results = int(top_k)
                    self.rag_config.similarity_threshold = float(similarity_threshold)
                    self.rag_config.temperature = float(temperature)
                    self.rag_config.max_tokens = int(max_tokens)
                    
                    return "✅ Configuration updated successfully!"
                except Exception as e:
                    return f"❌ Error updating configuration: {str(e)}"
            
            def search_documents(query):
                if not query.strip():
                    return {"error": "Please enter a search query"}
                
                results = self.search_similar(query, limit=5)
                return {
                    "query": query,
                    "results_count": len(results),
                    "results": [
                        {
                            "score": r["score"],
                            "text": r["text"][:200] + "..." if len(r["text"]) > 200 else r["text"],
                            "source": r["source"]
                        }
                        for r in results
                    ]
                }
            
            # Wire up the interface
            submit_btn.click(submit_message, inputs=[msg, chatbot], outputs=[msg, chatbot])
            msg.submit(submit_message, inputs=[msg, chatbot], outputs=[msg, chatbot])
            
            clear_btn.click(lambda: ([], ""), outputs=[chatbot, msg])
            
            add_docs_btn.click(
                self.add_documents_from_text,
                inputs=[doc_input, source_name],
                outputs=[add_result]
            )
            
            refresh_collection_btn.click(
                self.get_system_status,
                outputs=[collection_status]
            )
            
            search_btn.click(
                search_documents,
                inputs=[search_query],
                outputs=[search_results]
            )
            
            refresh_status_btn.click(
                self.get_system_status,
                outputs=[system_status_display]
            )
            
            refresh_metrics_btn.click(
                self.get_performance_metrics,
                outputs=[performance_metrics]
            )
            
            update_config_btn.click(
                update_configuration,
                inputs=[top_k_slider, similarity_threshold_slider, temperature_slider, max_tokens_slider],
                outputs=[config_result]
            )
            
            # Auto-refresh on load
            demo.load(self.get_system_status, outputs=[system_status_display])
            demo.load(self.get_performance_metrics, outputs=[performance_metrics])
            demo.load(self.get_system_status, outputs=[collection_status])
        
        return demo

def main():
    """
    Main function to launch the BMasterAI Qdrant RAG Gradio application
    """
    
    # Configure logging
    configure_logging(log_level=LogLevel.INFO)
    
    # Start monitoring
    monitor = get_monitor()
    monitor.start_monitoring()
    
    # Environment setup instructions
    required_vars = ["QDRANT_URL", "QDRANT_API_KEY", "OPENAI_API_KEY"]
    missing_vars = [var for var in required_vars if not os.getenv(var)]
    
    if missing_vars:
        print("\n" + "="*60)
        print("🔑 SETUP REQUIRED:")
        print("Missing environment variables:")
        for var in missing_vars:
            if "QDRANT" in var:
                print(f"  {var}='your-qdrant-{var.lower().split('_')[1]}'")
            else:
                print(f"  {var}='your-openai-api-key'")
        
        print("\nSet them with:")
        for var in missing_vars:
            print(f"export {var}='your-{var.lower().replace('_', '-')}'")
        
        print("\nOr create a .env file with:")
        for var in missing_vars:
            print(f"{var}=your-{var.lower().replace('_', '-')}")
        
        print("\n🌐 Get your keys from:")
        print("  Qdrant Cloud: https://cloud.qdrant.io/")
        print("  OpenAI: https://platform.openai.com/")
        print("="*60 + "\n")
        
        # Continue anyway for demo purposes
    
    # Configuration
    qdrant_config = QdrantConfig(
        url=os.getenv("QDRANT_URL", "https://your-cluster.qdrant.io"),
        api_key=os.getenv("QDRANT_API_KEY", ""),
        collection_name=os.getenv("QDRANT_COLLECTION", "bmasterai_rag_demo")
    )
    
    rag_config = RAGConfig(
        openai_api_key=os.getenv("OPENAI_API_KEY", ""),
        embedding_model=os.getenv("EMBEDDING_MODEL", "all-MiniLM-L6-v2"),
        llm_model=os.getenv("LLM_MODEL", "gpt-3.5-turbo"),
        top_k_results=int(os.getenv("TOP_K_RESULTS", "5"))
    )
    
    try:
        # Create and launch the application
        print("🚀 Initializing BMasterAI Qdrant RAG Gradio application...")
        app = BMasterAIQdrantRAGUI(qdrant_config, rag_config)
        demo = app.create_gradio_interface()
        
        # Launch configuration
        server_name = os.getenv("GRADIO_SERVER_NAME", "0.0.0.0")
        server_port = int(os.getenv("GRADIO_SERVER_PORT", "7860"))
        share = os.getenv("GRADIO_SHARE", "false").lower() == "true"
        
        print(f"🌐 Launching Gradio interface on {server_name}:{server_port}")
        print(f"📊 System Status: {app.system_status}")
        
        if not missing_vars:
            print("✅ All environment variables configured!")
        else:
            print("⚠️  Some environment variables missing - limited functionality")
        
        demo.launch(
            server_name=server_name,
            server_port=server_port,
            share=share,
            show_error=True,
            debug=False
        )
        
    except Exception as e:
        print(f"❌ Error launching application: {str(e)}")
        print("\n💡 Troubleshooting:")
        print("1. Check your environment variables")
        print("2. Verify Qdrant Cloud and OpenAI API keys")
        print("3. Ensure all dependencies are installed")
        print("4. Check network connectivity")
    
    finally:
        # Stop monitoring
        monitor.stop_monitoring()
        print("\n✅ BMasterAI Qdrant RAG Gradio application stopped!")

if __name__ == "__main__":
    main()


# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# Code generated by the Google Gen AI SDK generator DO NOT EDIT.

import datetime
import functools
import json
import logging
from typing import Any, Iterator, Optional, Union
from urllib.parse import urlencode

from google.genai import _api_module
from google.genai import _common
from google.genai._common import get_value_by_path as getv
from google.genai._common import set_value_by_path as setv
from google.genai.pagers import AsyncPager, Pager

from . import types


logger = logging.getLogger("vertexai_genai.sessionevents")

logger.setLevel(logging.INFO)


def _AppendAgentEngineSessionEventConfig_to_vertex(
    from_object: Union[dict[str, Any], object],
    parent_object: Optional[dict[str, Any]] = None,
) -> dict[str, Any]:
    to_object: dict[str, Any] = {}

    if getv(from_object, ["content"]) is not None:
        setv(parent_object, ["content"], getv(from_object, ["content"]))

    if getv(from_object, ["actions"]) is not None:
        setv(parent_object, ["actions"], getv(from_object, ["actions"]))

    if getv(from_object, ["error_code"]) is not None:
        setv(parent_object, ["errorCode"], getv(from_object, ["error_code"]))

    if getv(from_object, ["error_message"]) is not None:
        setv(parent_object, ["errorMessage"], getv(from_object, ["error_message"]))

    if getv(from_object, ["event_metadata"]) is not None:
        setv(parent_object, ["eventMetadata"], getv(from_object, ["event_metadata"]))

    return to_object


def _AppendAgentEngineSessionEventRequestParameters_to_vertex(
    from_object: Union[dict[str, Any], object],
    parent_object: Optional[dict[str, Any]] = None,
) -> dict[str, Any]:
    to_object: dict[str, Any] = {}
    if getv(from_object, ["name"]) is not None:
        setv(to_object, ["_url", "name"], getv(from_object, ["name"]))

    if getv(from_object, ["author"]) is not None:
        setv(to_object, ["author"], getv(from_object, ["author"]))

    if getv(from_object, ["invocation_id"]) is not None:
        setv(to_object, ["invocationId"], getv(from_object, ["invocation_id"]))

    if getv(from_object, ["timestamp"]) is not None:
        setv(to_object, ["timestamp"], getv(from_object, ["timestamp"]))

    if getv(from_object, ["config"]) is not None:
        setv(
            to_object,
            ["config"],
            _AppendAgentEngineSessionEventConfig_to_vertex(
                getv(from_object, ["config"]), to_object
            ),
        )

    return to_object


def _ListAgentEngineSessionEventsConfig_to_vertex(
    from_object: Union[dict[str, Any], object],
    parent_object: Optional[dict[str, Any]] = None,
) -> dict[str, Any]:
    to_object: dict[str, Any] = {}

    if getv(from_object, ["page_size"]) is not None:
        setv(parent_object, ["_query", "pageSize"], getv(from_object, ["page_size"]))

    if getv(from_object, ["page_token"]) is not None:
        setv(parent_object, ["_query", "pageToken"], getv(from_object, ["page_token"]))

    if getv(from_object, ["filter"]) is not None:
        setv(parent_object, ["_query", "filter"], getv(from_object, ["filter"]))

    return to_object


def _ListAgentEngineSessionEventsRequestParameters_to_vertex(
    from_object: Union[dict[str, Any], object],
    parent_object: Optional[dict[str, Any]] = None,
) -> dict[str, Any]:
    to_object: dict[str, Any] = {}
    if getv(from_object, ["name"]) is not None:
        setv(to_object, ["_url", "name"], getv(from_object, ["name"]))

    if getv(from_object, ["config"]) is not None:
        setv(
            to_object,
            ["config"],
            _ListAgentEngineSessionEventsConfig_to_vertex(
                getv(from_object, ["config"]), to_object
            ),
        )

    return to_object


class SessionEvents(_api_module.BaseModule):

    def append(
        self,
        *,
        name: str,
        author: str,
        invocation_id: str,
        timestamp: datetime.datetime,
        config: Optional[types.AppendAgentEngineSessionEventConfigOrDict] = None,
    ) -> types.AppendAgentEngineSessionEventResponse:
        """
        Appends Agent Engine session event.

        Args:
            name (str): Required. The name of the Agent Engine session to append the event to. Format:
                `projects/{project}/locations/{location}/reasoningEngines/{resource_id}/sessions/{session_id}`.
            author (str): Required. The author of the Agent Engine session event.
            invocation_id (str): Required. The invocation ID of the Agent Engine session event.
            timestamp (datetime.datetime): Required. The timestamp of the Agent Engine session event.
            config (AppendAgentEngineSessionEventConfig):
                Optional. Additional configurations for appending the Agent Engine session event.

        Returns:
            AppendAgentEngineSessionEventResponse: The requested Agent Engine session event.

        """

        parameter_model = types._AppendAgentEngineSessionEventRequestParameters(
            name=name,
            author=author,
            invocation_id=invocation_id,
            timestamp=timestamp,
            config=config,
        )

        request_url_dict: Optional[dict[str, str]]
        if not self._api_client.vertexai:
            raise ValueError("This method is only supported in the Vertex AI client.")
        else:
            request_dict = _AppendAgentEngineSessionEventRequestParameters_to_vertex(
                parameter_model
            )
            request_url_dict = request_dict.get("_url")
            if request_url_dict:
                path = "{name}:appendEvent".format_map(request_url_dict)
            else:
                path = "{name}:appendEvent"

        query_params = request_dict.get("_query")
        if query_params:
            path = f"{path}?{urlencode(query_params)}"
        # TODO: remove the hack that pops config.
        request_dict.pop("config", None)

        http_options: Optional[types.HttpOptions] = None
        if (
            parameter_model.config is not None
            and parameter_model.config.http_options is not None
        ):
            http_options = parameter_model.config.http_options

        request_dict = _common.convert_to_dict(request_dict)
        request_dict = _common.encode_unserializable_types(request_dict)

        response = self._api_client.request("post", path, request_dict, http_options)

        response_dict = {} if not response.body else json.loads(response.body)

        return_value = types.AppendAgentEngineSessionEventResponse._from_response(
            response=response_dict, kwargs=parameter_model.model_dump()
        )

        self._api_client._verify_response(return_value)
        return return_value

    def _list(
        self,
        *,
        name: str,
        config: Optional[types.ListAgentEngineSessionEventsConfigOrDict] = None,
    ) -> types.ListAgentEngineSessionEventsResponse:
        """
        Lists Agent Engine session events.

        Args:
            name (str): Required. The name of the Agent Engine session to list events for. Format:
                `projects/{project}/locations/{location}/reasoningEngines/{resource_id}/sessions/{session_id}`.
            config (ListAgentEngineSessionEventsConfig):
                Optional. Additional configurations for listing the Agent Engine session events.

        Returns:
            ListAgentEngineSessionEventsResponse: The requested Agent Engine session events.

        """

        parameter_model = types._ListAgentEngineSessionEventsRequestParameters(
            name=name,
            config=config,
        )

        request_url_dict: Optional[dict[str, str]]
        if not self._api_client.vertexai:
            raise ValueError("This method is only supported in the Vertex AI client.")
        else:
            request_dict = _ListAgentEngineSessionEventsRequestParameters_to_vertex(
                parameter_model
            )
            request_url_dict = request_dict.get("_url")
            if request_url_dict:
                path = "{name}/events".format_map(request_url_dict)
            else:
                path = "{name}/events"

        query_params = request_dict.get("_query")
        if query_params:
            path = f"{path}?{urlencode(query_params)}"
        # TODO: remove the hack that pops config.
        request_dict.pop("config", None)

        http_options: Optional[types.HttpOptions] = None
        if (
            parameter_model.config is not None
            and parameter_model.config.http_options is not None
        ):
            http_options = parameter_model.config.http_options

        request_dict = _common.convert_to_dict(request_dict)
        request_dict = _common.encode_unserializable_types(request_dict)

        response = self._api_client.request("get", path, request_dict, http_options)

        response_dict = {} if not response.body else json.loads(response.body)

        return_value = types.ListAgentEngineSessionEventsResponse._from_response(
            response=response_dict, kwargs=parameter_model.model_dump()
        )

        self._api_client._verify_response(return_value)
        return return_value

    def list(
        self,
        *,
        name: str,
        config: Optional[types.ListAgentEngineSessionEventsConfigOrDict] = None,
    ) -> Iterator[types.SessionEvent]:
        """Lists Agent Engine session events.

        Args:
            name (str): Required. The name of the agent engine to list session
                events for.
            config (ListAgentEngineSessionEventsConfig): Optional. The configuration
                for the session events to list. Currently, the `filter` field in
                `config` only supports filtering by `timestamp`. The timestamp
                value must be enclosed in double quotes and include the time zone
                information. For example:
                `config={'filter': 'timestamp>="2025-08-07T19:44:38.4Z"'}`.

        Returns:
            Iterator[SessionEvent]: An iterable of session events.
        """

        return Pager(
            "session_events",
            functools.partial(self._list, name=name),
            self._list(name=name, config=config),
            config,
        )


class AsyncSessionEvents(_api_module.BaseModule):

    async def append(
        self,
        *,
        name: str,
        author: str,
        invocation_id: str,
        timestamp: datetime.datetime,
        config: Optional[types.AppendAgentEngineSessionEventConfigOrDict] = None,
    ) -> types.AppendAgentEngineSessionEventResponse:
        """
        Appends Agent Engine session event.

        Args:
            name (str): Required. The name of the Agent Engine session to append the event to. Format:
                `projects/{project}/locations/{location}/reasoningEngines/{resource_id}/sessions/{session_id}`.
            author (str): Required. The author of the Agent Engine session event.
            invocation_id (str): Required. The invocation ID of the Agent Engine session event.
            timestamp (datetime.datetime): Required. The timestamp of the Agent Engine session event.
            config (AppendAgentEngineSessionEventConfig):
                Optional. Additional configurations for appending the Agent Engine session event.

        Returns:
            AppendAgentEngineSessionEventResponse: The requested Agent Engine session event.

        """

        parameter_model = types._AppendAgentEngineSessionEventRequestParameters(
            name=name,
            author=author,
            invocation_id=invocation_id,
            timestamp=timestamp,
            config=config,
        )

        request_url_dict: Optional[dict[str, str]]
        if not self._api_client.vertexai:
            raise ValueError("This method is only supported in the Vertex AI client.")
        else:
            request_dict = _AppendAgentEngineSessionEventRequestParameters_to_vertex(
                parameter_model
            )
            request_url_dict = request_dict.get("_url")
            if request_url_dict:
                path = "{name}:appendEvent".format_map(request_url_dict)
            else:
                path = "{name}:appendEvent"

        query_params = request_dict.get("_query")
        if query_params:
            path = f"{path}?{urlencode(query_params)}"
        # TODO: remove the hack that pops config.
        request_dict.pop("config", None)

        http_options: Optional[types.HttpOptions] = None
        if (
            parameter_model.config is not None
            and parameter_model.config.http_options is not None
        ):
            http_options = parameter_model.config.http_options

        request_dict = _common.convert_to_dict(request_dict)
        request_dict = _common.encode_unserializable_types(request_dict)

        response = await self._api_client.async_request(
            "post", path, request_dict, http_options
        )

        response_dict = {} if not response.body else json.loads(response.body)

        return_value = types.AppendAgentEngineSessionEventResponse._from_response(
            response=response_dict, kwargs=parameter_model.model_dump()
        )

        self._api_client._verify_response(return_value)
        return return_value

    async def _list(
        self,
        *,
        name: str,
        config: Optional[types.ListAgentEngineSessionEventsConfigOrDict] = None,
    ) -> types.ListAgentEngineSessionEventsResponse:
        """
        Lists Agent Engine session events.

        Args:
            name (str): Required. The name of the Agent Engine session to list events for. Format:
                `projects/{project}/locations/{location}/reasoningEngines/{resource_id}/sessions/{session_id}`.
            config (ListAgentEngineSessionEventsConfig):
                Optional. Additional configurations for listing the Agent Engine session events.

        Returns:
            ListAgentEngineSessionEventsResponse: The requested Agent Engine session events.

        """

        parameter_model = types._ListAgentEngineSessionEventsRequestParameters(
            name=name,
            config=config,
        )

        request_url_dict: Optional[dict[str, str]]
        if not self._api_client.vertexai:
            raise ValueError("This method is only supported in the Vertex AI client.")
        else:
            request_dict = _ListAgentEngineSessionEventsRequestParameters_to_vertex(
                parameter_model
            )
            request_url_dict = request_dict.get("_url")
            if request_url_dict:
                path = "{name}/events".format_map(request_url_dict)
            else:
                path = "{name}/events"

        query_params = request_dict.get("_query")
        if query_params:
            path = f"{path}?{urlencode(query_params)}"
        # TODO: remove the hack that pops config.
        request_dict.pop("config", None)

        http_options: Optional[types.HttpOptions] = None
        if (
            parameter_model.config is not None
            and parameter_model.config.http_options is not None
        ):
            http_options = parameter_model.config.http_options

        request_dict = _common.convert_to_dict(request_dict)
        request_dict = _common.encode_unserializable_types(request_dict)

        response = await self._api_client.async_request(
            "get", path, request_dict, http_options
        )

        response_dict = {} if not response.body else json.loads(response.body)

        return_value = types.ListAgentEngineSessionEventsResponse._from_response(
            response=response_dict, kwargs=parameter_model.model_dump()
        )

        self._api_client._verify_response(return_value)
        return return_value

    async def list(
        self,
        *,
        name: str,
        config: Optional[types.ListAgentEngineSessionEventsConfigOrDict] = None,
    ) -> AsyncPager[types.SessionEvent]:
        """Lists Agent Engine session events.

        Args:
            name (str): Required. The name of the agent engine to list session
                events for.
            config (ListAgentEngineSessionEventsConfig): Optional. The configuration
                for the session events to list. Currently, the `filter` field in
                `config` only supports filtering by `timestamp`. The timestamp
                value must be enclosed in double quotes and include the time zone
                information. For example:
                `config={'filter': 'timestamp>="2025-08-07T19:44:38.4Z"'}`.

        Returns:
            AsyncPager[SessionEvent]: An async pager of session events.
        """

        return AsyncPager(
            "session_events",
            functools.partial(self._list, name=name),
            await self._list(name=name, config=config),
            config,
        )



# -*- coding: utf-8 -*-
# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations

from typing import MutableMapping, MutableSequence

import proto  # type: ignore

from google.cloud.aiplatform_v1.types import featurestore_online_service
from google.protobuf import struct_pb2  # type: ignore
from google.protobuf import timestamp_pb2  # type: ignore
from google.rpc import status_pb2  # type: ignore


__protobuf__ = proto.module(
    package="google.cloud.aiplatform.v1",
    manifest={
        "FeatureViewDataFormat",
        "FeatureViewDataKey",
        "FetchFeatureValuesRequest",
        "FetchFeatureValuesResponse",
        "NearestNeighborQuery",
        "SearchNearestEntitiesRequest",
        "NearestNeighbors",
        "SearchNearestEntitiesResponse",
        "FeatureViewDirectWriteRequest",
        "FeatureViewDirectWriteResponse",
        "GenerateFetchAccessTokenRequest",
        "GenerateFetchAccessTokenResponse",
    },
)


class FeatureViewDataFormat(proto.Enum):
    r"""Format of the data in the Feature View.

    Values:
        FEATURE_VIEW_DATA_FORMAT_UNSPECIFIED (0):
            Not set. Will be treated as the KeyValue
            format.
        KEY_VALUE (1):
            Return response data in key-value format.
        PROTO_STRUCT (2):
            Return response data in proto Struct format.
    """

    FEATURE_VIEW_DATA_FORMAT_UNSPECIFIED = 0
    KEY_VALUE = 1
    PROTO_STRUCT = 2


class FeatureViewDataKey(proto.Message):
    r"""Lookup key for a feature view.

    This message has `oneof`_ fields (mutually exclusive fields).
    For each oneof, at most one member field can be set at the same time.
    Setting any member of the oneof automatically clears all other
    members.

    .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

    Attributes:
        key (str):
            String key to use for lookup.

            This field is a member of `oneof`_ ``key_oneof``.
        composite_key (google.cloud.aiplatform_v1.types.FeatureViewDataKey.CompositeKey):
            The actual Entity ID will be composed from
            this struct. This should match with the way ID
            is defined in the FeatureView spec.

            This field is a member of `oneof`_ ``key_oneof``.
    """

    class CompositeKey(proto.Message):
        r"""ID that is comprised from several parts (columns).

        Attributes:
            parts (MutableSequence[str]):
                Parts to construct Entity ID. Should match
                with the same ID columns as defined in
                FeatureView in the same order.
        """

        parts: MutableSequence[str] = proto.RepeatedField(
            proto.STRING,
            number=1,
        )

    key: str = proto.Field(
        proto.STRING,
        number=1,
        oneof="key_oneof",
    )
    composite_key: CompositeKey = proto.Field(
        proto.MESSAGE,
        number=2,
        oneof="key_oneof",
        message=CompositeKey,
    )


class FetchFeatureValuesRequest(proto.Message):
    r"""Request message for
    [FeatureOnlineStoreService.FetchFeatureValues][google.cloud.aiplatform.v1.FeatureOnlineStoreService.FetchFeatureValues].
    All the features under the requested feature view will be returned.

    Attributes:
        feature_view (str):
            Required. FeatureView resource format
            ``projects/{project}/locations/{location}/featureOnlineStores/{featureOnlineStore}/featureViews/{featureView}``
        data_key (google.cloud.aiplatform_v1.types.FeatureViewDataKey):
            Optional. The request key to fetch feature
            values for.
        data_format (google.cloud.aiplatform_v1.types.FeatureViewDataFormat):
            Optional. Response data format. If not set,
            [FeatureViewDataFormat.KEY_VALUE][google.cloud.aiplatform.v1.FeatureViewDataFormat.KEY_VALUE]
            will be used.
    """

    feature_view: str = proto.Field(
        proto.STRING,
        number=1,
    )
    data_key: "FeatureViewDataKey" = proto.Field(
        proto.MESSAGE,
        number=6,
        message="FeatureViewDataKey",
    )
    data_format: "FeatureViewDataFormat" = proto.Field(
        proto.ENUM,
        number=7,
        enum="FeatureViewDataFormat",
    )


class FetchFeatureValuesResponse(proto.Message):
    r"""Response message for
    [FeatureOnlineStoreService.FetchFeatureValues][google.cloud.aiplatform.v1.FeatureOnlineStoreService.FetchFeatureValues]

    This message has `oneof`_ fields (mutually exclusive fields).
    For each oneof, at most one member field can be set at the same time.
    Setting any member of the oneof automatically clears all other
    members.

    .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

    Attributes:
        key_values (google.cloud.aiplatform_v1.types.FetchFeatureValuesResponse.FeatureNameValuePairList):
            Feature values in KeyValue format.

            This field is a member of `oneof`_ ``format``.
        proto_struct (google.protobuf.struct_pb2.Struct):
            Feature values in proto Struct format.

            This field is a member of `oneof`_ ``format``.
        data_key (google.cloud.aiplatform_v1.types.FeatureViewDataKey):
            The data key associated with this response. Will only be
            populated for
            [FeatureOnlineStoreService.StreamingFetchFeatureValues][]
            RPCs.
    """

    class FeatureNameValuePairList(proto.Message):
        r"""Response structure in the format of key (feature name) and
        (feature) value pair.

        Attributes:
            features (MutableSequence[google.cloud.aiplatform_v1.types.FetchFeatureValuesResponse.FeatureNameValuePairList.FeatureNameValuePair]):
                List of feature names and values.
        """

        class FeatureNameValuePair(proto.Message):
            r"""Feature name & value pair.

            .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

            Attributes:
                value (google.cloud.aiplatform_v1.types.FeatureValue):
                    Feature value.

                    This field is a member of `oneof`_ ``data``.
                name (str):
                    Feature short name.
            """

            value: featurestore_online_service.FeatureValue = proto.Field(
                proto.MESSAGE,
                number=2,
                oneof="data",
                message=featurestore_online_service.FeatureValue,
            )
            name: str = proto.Field(
                proto.STRING,
                number=1,
            )

        features: MutableSequence[
            "FetchFeatureValuesResponse.FeatureNameValuePairList.FeatureNameValuePair"
        ] = proto.RepeatedField(
            proto.MESSAGE,
            number=1,
            message="FetchFeatureValuesResponse.FeatureNameValuePairList.FeatureNameValuePair",
        )

    key_values: FeatureNameValuePairList = proto.Field(
        proto.MESSAGE,
        number=3,
        oneof="format",
        message=FeatureNameValuePairList,
    )
    proto_struct: struct_pb2.Struct = proto.Field(
        proto.MESSAGE,
        number=2,
        oneof="format",
        message=struct_pb2.Struct,
    )
    data_key: "FeatureViewDataKey" = proto.Field(
        proto.MESSAGE,
        number=4,
        message="FeatureViewDataKey",
    )


class NearestNeighborQuery(proto.Message):
    r"""A query to find a number of similar entities.

    This message has `oneof`_ fields (mutually exclusive fields).
    For each oneof, at most one member field can be set at the same time.
    Setting any member of the oneof automatically clears all other
    members.

    .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

    Attributes:
        entity_id (str):
            Optional. The entity id whose similar entities should be
            searched for. If embedding is set, search will use embedding
            instead of entity_id.

            This field is a member of `oneof`_ ``instance``.
        embedding (google.cloud.aiplatform_v1.types.NearestNeighborQuery.Embedding):
            Optional. The embedding vector that be used
            for similar search.

            This field is a member of `oneof`_ ``instance``.
        neighbor_count (int):
            Optional. The number of similar entities to
            be retrieved from feature view for each query.
        string_filters (MutableSequence[google.cloud.aiplatform_v1.types.NearestNeighborQuery.StringFilter]):
            Optional. The list of string filters.
        numeric_filters (MutableSequence[google.cloud.aiplatform_v1.types.NearestNeighborQuery.NumericFilter]):
            Optional. The list of numeric filters.
        per_crowding_attribute_neighbor_count (int):
            Optional. Crowding is a constraint on a neighbor list
            produced by nearest neighbor search requiring that no more
            than sper_crowding_attribute_neighbor_count of the k
            neighbors returned have the same value of
            crowding_attribute. It's used for improving result
            diversity.
        parameters (google.cloud.aiplatform_v1.types.NearestNeighborQuery.Parameters):
            Optional. Parameters that can be set to tune
            query on the fly.
    """

    class Embedding(proto.Message):
        r"""The embedding vector.

        Attributes:
            value (MutableSequence[float]):
                Optional. Individual value in the embedding.
        """

        value: MutableSequence[float] = proto.RepeatedField(
            proto.FLOAT,
            number=1,
        )

    class StringFilter(proto.Message):
        r"""String filter is used to search a subset of the entities by using
        boolean rules on string columns. For example: if a query specifies
        string filter with 'name = color, allow_tokens = {red, blue},
        deny_tokens = {purple}',' then that query will match entities that
        are red or blue, but if those points are also purple, then they will
        be excluded even if they are red/blue. Only string filter is
        supported for now, numeric filter will be supported in the near
        future.

        Attributes:
            name (str):
                Required. Column names in BigQuery that used
                as filters.
            allow_tokens (MutableSequence[str]):
                Optional. The allowed tokens.
            deny_tokens (MutableSequence[str]):
                Optional. The denied tokens.
        """

        name: str = proto.Field(
            proto.STRING,
            number=1,
        )
        allow_tokens: MutableSequence[str] = proto.RepeatedField(
            proto.STRING,
            number=2,
        )
        deny_tokens: MutableSequence[str] = proto.RepeatedField(
            proto.STRING,
            number=3,
        )

    class NumericFilter(proto.Message):
        r"""Numeric filter is used to search a subset of the entities by using
        boolean rules on numeric columns. For example: Database Point 0:
        {name: "a" value_int: 42} {name: "b" value_float: 1.0} Database
        Point 1: {name: "a" value_int: 10} {name: "b" value_float: 2.0}
        Database Point 2: {name: "a" value_int: -1} {name: "b" value_float:
        3.0} Query: {name: "a" value_int: 12 operator: LESS} // Matches
        Point 1, 2 {name: "b" value_float: 2.0 operator: EQUAL} // Matches
        Point 1

        This message has `oneof`_ fields (mutually exclusive fields).
        For each oneof, at most one member field can be set at the same time.
        Setting any member of the oneof automatically clears all other
        members.

        .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

        Attributes:
            value_int (int):
                int value type.

                This field is a member of `oneof`_ ``Value``.
            value_float (float):
                float value type.

                This field is a member of `oneof`_ ``Value``.
            value_double (float):
                double value type.

                This field is a member of `oneof`_ ``Value``.
            name (str):
                Required. Column name in BigQuery that used
                as filters.
            op (google.cloud.aiplatform_v1.types.NearestNeighborQuery.NumericFilter.Operator):
                Optional. This MUST be specified for queries
                and must NOT be specified for database points.

                This field is a member of `oneof`_ ``_op``.
        """

        class Operator(proto.Enum):
            r"""Datapoints for which Operator is true relative to the query's
            Value field will be allowlisted.

            Values:
                OPERATOR_UNSPECIFIED (0):
                    Unspecified operator.
                LESS (1):
                    Entities are eligible if their value is < the
                    query's.
                LESS_EQUAL (2):
                    Entities are eligible if their value is <=
                    the query's.
                EQUAL (3):
                    Entities are eligible if their value is ==
                    the query's.
                GREATER_EQUAL (4):
                    Entities are eligible if their value is >=
                    the query's.
                GREATER (5):
                    Entities are eligible if their value is > the
                    query's.
                NOT_EQUAL (6):
                    Entities are eligible if their value is !=
                    the query's.
            """

            OPERATOR_UNSPECIFIED = 0
            LESS = 1
            LESS_EQUAL = 2
            EQUAL = 3
            GREATER_EQUAL = 4
            GREATER = 5
            NOT_EQUAL = 6

        value_int: int = proto.Field(
            proto.INT64,
            number=2,
            oneof="Value",
        )
        value_float: float = proto.Field(
            proto.FLOAT,
            number=3,
            oneof="Value",
        )
        value_double: float = proto.Field(
            proto.DOUBLE,
            number=4,
            oneof="Value",
        )
        name: str = proto.Field(
            proto.STRING,
            number=1,
        )
        op: "NearestNeighborQuery.NumericFilter.Operator" = proto.Field(
            proto.ENUM,
            number=5,
            optional=True,
            enum="NearestNeighborQuery.NumericFilter.Operator",
        )

    class Parameters(proto.Message):
        r"""Parameters that can be overrided in each query to tune query
        latency and recall.

        Attributes:
            approximate_neighbor_candidates (int):
                Optional. The number of neighbors to find via approximate
                search before exact reordering is performed; if set, this
                value must be > neighbor_count.
            leaf_nodes_search_fraction (float):
                Optional. The fraction of the number of
                leaves to search, set at query time allows user
                to tune search performance. This value increase
                result in both search accuracy and latency
                increase. The value should be between 0.0 and
                1.0.
        """

        approximate_neighbor_candidates: int = proto.Field(
            proto.INT32,
            number=1,
        )
        leaf_nodes_search_fraction: float = proto.Field(
            proto.DOUBLE,
            number=2,
        )

    entity_id: str = proto.Field(
        proto.STRING,
        number=1,
        oneof="instance",
    )
    embedding: Embedding = proto.Field(
        proto.MESSAGE,
        number=2,
        oneof="instance",
        message=Embedding,
    )
    neighbor_count: int = proto.Field(
        proto.INT32,
        number=3,
    )
    string_filters: MutableSequence[StringFilter] = proto.RepeatedField(
        proto.MESSAGE,
        number=4,
        message=StringFilter,
    )
    numeric_filters: MutableSequence[NumericFilter] = proto.RepeatedField(
        proto.MESSAGE,
        number=8,
        message=NumericFilter,
    )
    per_crowding_attribute_neighbor_count: int = proto.Field(
        proto.INT32,
        number=5,
    )
    parameters: Parameters = proto.Field(
        proto.MESSAGE,
        number=7,
        message=Parameters,
    )


class SearchNearestEntitiesRequest(proto.Message):
    r"""The request message for
    [FeatureOnlineStoreService.SearchNearestEntities][google.cloud.aiplatform.v1.FeatureOnlineStoreService.SearchNearestEntities].

    Attributes:
        feature_view (str):
            Required. FeatureView resource format
            ``projects/{project}/locations/{location}/featureOnlineStores/{featureOnlineStore}/featureViews/{featureView}``
        query (google.cloud.aiplatform_v1.types.NearestNeighborQuery):
            Required. The query.
        return_full_entity (bool):
            Optional. If set to true, the full entities
            (including all vector values and metadata) of
            the nearest neighbors are returned; otherwise
            only entity id of the nearest neighbors will be
            returned. Note that returning full entities will
            significantly increase the latency and cost of
            the query.
    """

    feature_view: str = proto.Field(
        proto.STRING,
        number=1,
    )
    query: "NearestNeighborQuery" = proto.Field(
        proto.MESSAGE,
        number=2,
        message="NearestNeighborQuery",
    )
    return_full_entity: bool = proto.Field(
        proto.BOOL,
        number=3,
    )


class NearestNeighbors(proto.Message):
    r"""Nearest neighbors for one query.

    Attributes:
        neighbors (MutableSequence[google.cloud.aiplatform_v1.types.NearestNeighbors.Neighbor]):
            All its neighbors.
    """

    class Neighbor(proto.Message):
        r"""A neighbor of the query vector.

        Attributes:
            entity_id (str):
                The id of the similar entity.
            distance (float):
                The distance between the neighbor and the
                query vector.
            entity_key_values (google.cloud.aiplatform_v1.types.FetchFeatureValuesResponse):
                The attributes of the neighbor, e.g. filters, crowding and
                metadata Note that full entities are returned only when
                "return_full_entity" is set to true. Otherwise, only the
                "entity_id" and "distance" fields are populated.
        """

        entity_id: str = proto.Field(
            proto.STRING,
            number=1,
        )
        distance: float = proto.Field(
            proto.DOUBLE,
            number=2,
        )
        entity_key_values: "FetchFeatureValuesResponse" = proto.Field(
            proto.MESSAGE,
            number=3,
            message="FetchFeatureValuesResponse",
        )

    neighbors: MutableSequence[Neighbor] = proto.RepeatedField(
        proto.MESSAGE,
        number=1,
        message=Neighbor,
    )


class SearchNearestEntitiesResponse(proto.Message):
    r"""Response message for
    [FeatureOnlineStoreService.SearchNearestEntities][google.cloud.aiplatform.v1.FeatureOnlineStoreService.SearchNearestEntities]

    Attributes:
        nearest_neighbors (google.cloud.aiplatform_v1.types.NearestNeighbors):
            The nearest neighbors of the query entity.
    """

    nearest_neighbors: "NearestNeighbors" = proto.Field(
        proto.MESSAGE,
        number=1,
        message="NearestNeighbors",
    )


class FeatureViewDirectWriteRequest(proto.Message):
    r"""Request message for
    [FeatureOnlineStoreService.FeatureViewDirectWrite][google.cloud.aiplatform.v1.FeatureOnlineStoreService.FeatureViewDirectWrite].

    Attributes:
        feature_view (str):
            FeatureView resource format
            ``projects/{project}/locations/{location}/featureOnlineStores/{featureOnlineStore}/featureViews/{featureView}``
        data_key_and_feature_values (MutableSequence[google.cloud.aiplatform_v1.types.FeatureViewDirectWriteRequest.DataKeyAndFeatureValues]):
            Required. The data keys and associated
            feature values.
    """

    class DataKeyAndFeatureValues(proto.Message):
        r"""A data key and associated feature values to write to the
        feature view.

        Attributes:
            data_key (google.cloud.aiplatform_v1.types.FeatureViewDataKey):
                The data key.
            features (MutableSequence[google.cloud.aiplatform_v1.types.FeatureViewDirectWriteRequest.DataKeyAndFeatureValues.Feature]):
                List of features to write.
        """

        class Feature(proto.Message):
            r"""Feature name & value pair.

            .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

            Attributes:
                value (google.cloud.aiplatform_v1.types.FeatureValue):
                    Feature value. A user provided timestamp may be set in the
                    ``FeatureValue.metadata.generate_time`` field.

                    This field is a member of `oneof`_ ``data_oneof``.
                name (str):
                    Feature short name.
            """

            value: featurestore_online_service.FeatureValue = proto.Field(
                proto.MESSAGE,
                number=3,
                oneof="data_oneof",
                message=featurestore_online_service.FeatureValue,
            )
            name: str = proto.Field(
                proto.STRING,
                number=1,
            )

        data_key: "FeatureViewDataKey" = proto.Field(
            proto.MESSAGE,
            number=1,
            message="FeatureViewDataKey",
        )
        features: MutableSequence[
            "FeatureViewDirectWriteRequest.DataKeyAndFeatureValues.Feature"
        ] = proto.RepeatedField(
            proto.MESSAGE,
            number=2,
            message="FeatureViewDirectWriteRequest.DataKeyAndFeatureValues.Feature",
        )

    feature_view: str = proto.Field(
        proto.STRING,
        number=1,
    )
    data_key_and_feature_values: MutableSequence[DataKeyAndFeatureValues] = (
        proto.RepeatedField(
            proto.MESSAGE,
            number=2,
            message=DataKeyAndFeatureValues,
        )
    )


class FeatureViewDirectWriteResponse(proto.Message):
    r"""Response message for
    [FeatureOnlineStoreService.FeatureViewDirectWrite][google.cloud.aiplatform.v1.FeatureOnlineStoreService.FeatureViewDirectWrite].

    Attributes:
        status (google.rpc.status_pb2.Status):
            Response status for the keys listed in
            [FeatureViewDirectWriteResponse.write_responses][google.cloud.aiplatform.v1.FeatureViewDirectWriteResponse.write_responses].

            The error only applies to the listed data keys - the stream
            will remain open for further
            [FeatureOnlineStoreService.FeatureViewDirectWriteRequest][]
            requests.

            Partial failures (e.g. if the first 10 keys of a request
            fail, but the rest succeed) from a single request may result
            in multiple responses - there will be one response for the
            successful request keys and one response for the failing
            request keys.
        write_responses (MutableSequence[google.cloud.aiplatform_v1.types.FeatureViewDirectWriteResponse.WriteResponse]):
            Details about write for each key. If status is not OK,
            [WriteResponse.data_key][google.cloud.aiplatform.v1.FeatureViewDirectWriteResponse.WriteResponse.data_key]
            will have the key with error, but
            [WriteResponse.online_store_write_time][google.cloud.aiplatform.v1.FeatureViewDirectWriteResponse.WriteResponse.online_store_write_time]
            will not be present.
    """

    class WriteResponse(proto.Message):
        r"""Details about the write for each key.

        Attributes:
            data_key (google.cloud.aiplatform_v1.types.FeatureViewDataKey):
                What key is this write response associated
                with.
            online_store_write_time (google.protobuf.timestamp_pb2.Timestamp):
                When the feature values were written to the online store. If
                [FeatureViewDirectWriteResponse.status][google.cloud.aiplatform.v1.FeatureViewDirectWriteResponse.status]
                is not OK, this field is not populated.
        """

        data_key: "FeatureViewDataKey" = proto.Field(
            proto.MESSAGE,
            number=1,
            message="FeatureViewDataKey",
        )
        online_store_write_time: timestamp_pb2.Timestamp = proto.Field(
            proto.MESSAGE,
            number=2,
            message=timestamp_pb2.Timestamp,
        )

    status: status_pb2.Status = proto.Field(
        proto.MESSAGE,
        number=1,
        message=status_pb2.Status,
    )
    write_responses: MutableSequence[WriteResponse] = proto.RepeatedField(
        proto.MESSAGE,
        number=2,
        message=WriteResponse,
    )


class GenerateFetchAccessTokenRequest(proto.Message):
    r"""Request message for
    [FeatureOnlineStoreService.GenerateFetchAccessToken][google.cloud.aiplatform.v1.FeatureOnlineStoreService.GenerateFetchAccessToken].

    Attributes:
        feature_view (str):
            FeatureView resource format
            ``projects/{project}/locations/{location}/featureOnlineStores/{featureOnlineStore}/featureViews/{featureView}``
    """

    feature_view: str = proto.Field(
        proto.STRING,
        number=1,
    )


class GenerateFetchAccessTokenResponse(proto.Message):
    r"""Response message for
    [FeatureOnlineStoreService.GenerateFetchAccessToken][google.cloud.aiplatform.v1.FeatureOnlineStoreService.GenerateFetchAccessToken].

    Attributes:
        access_token (str):
            The OAuth 2.0 access token.
        expire_time (google.protobuf.timestamp_pb2.Timestamp):
            Token expiration time. This is always set
    """

    access_token: str = proto.Field(
        proto.STRING,
        number=1,
    )
    expire_time: timestamp_pb2.Timestamp = proto.Field(
        proto.MESSAGE,
        number=2,
        message=timestamp_pb2.Timestamp,
    )


__all__ = tuple(sorted(__protobuf__.manifest))



"""
LinkedIn Chain of Thought Wellness Advisor

This module implements a wellness advisor that analyzes LinkedIn profiles to provide
personalized stress reduction and happiness suggestions using transparent chain of 
thought reasoning with BMasterAI.
"""

import time
import json
import os
import sys
import requests
import re
from typing import Dict, List, Any, Optional, Callable
from datetime import datetime
import google.generativeai as genai

# Import the local mock data API for LinkedIn data access
from data_api import ApiClient

# Import BMasterAI components for full transparency
try:
    from bmasterai import (
        configure_logging, get_logger, get_monitor,
        LogLevel
    )
    BMASTERAI_AVAILABLE = True
    
    # Try importing reasoning components (may not be available in all versions)
    try:
        from bmasterai import ReasoningSession, ChainOfThought, with_reasoning_logging, log_reasoning
        REASONING_AVAILABLE = True
    except ImportError:
        REASONING_AVAILABLE = False
        # Mock reasoning components not needed since ChainOfThought is disabled
        ReasoningSession = None
        ChainOfThought = None
        def with_reasoning_logging(func):
            return func
        def log_reasoning(agent_id, task, thoughts, conclusion):
            pass
        
except ImportError:
    BMASTERAI_AVAILABLE = False
    REASONING_AVAILABLE = False
    print("BMasterAI library not found. Please install it with: pip install bmasterai")
    
    # Create mock classes for when bmasterai is not available
    class MockLogger:
        def info(self, msg): print(f"INFO: {msg}")
        def warning(self, msg): print(f"WARNING: {msg}")
        def error(self, msg): print(f"ERROR: {msg}")
    
    def get_logger(name):
        return MockLogger()
    
    def get_monitor():
        return None
    
    def configure_logging(**kwargs):
        pass
    
    class LogLevel:
        INFO = "INFO"
        WARNING = "WARNING"
        ERROR = "ERROR"


class LinkedInWellnessAgent:
    """
        Chain of thought wellness advisor powered by Gemini + BMasterAI + LinkedIn APIs
        Provides transparent reasoning for personalized stress reduction and happiness suggestions
    """
    
    def __init__(self, agent_id: str, gemini_api_key: str):
        self.agent_id = agent_id
        self.gemini_api_key = gemini_api_key
        self.model_name = "gemini-2.0-flash-exp"
        
        # Configure Gemini
        genai.configure(api_key=gemini_api_key)
        self.model = genai.GenerativeModel(self.model_name)
        
        # Initialize LinkedIn API client
        self.linkedin_client = ApiClient()
        
        # Get BMasterAI logger and monitor if available
        if BMASTERAI_AVAILABLE:
            self.logger = get_logger()
            self.monitor = get_monitor()
            self.monitor.start_monitoring()
        else:
            self.logger = None
            self.monitor = None
        
        # Callback for real-time UI updates
        self.update_callback: Optional[Callable] = None
    
    def set_update_callback(self, callback: Callable):
        """Set callback function for real-time UI updates"""
        self.update_callback = callback
    
    def _send_reasoning_update(self, step_type: str, reasoning: str, conclusion: str = "", data: Any = None, confidence: float = 0.8):
        """Send reasoning update to UI callback and BMasterAI logging"""
        
        update_data = {
        'type': step_type,
        'reasoning': reasoning,
        'conclusion': conclusion,
        'data': data,
        'confidence': confidence,
        'timestamp': datetime.now().isoformat()
        }
        
        # Log to BMasterAI if available
        if BMASTERAI_AVAILABLE and self.logger:
            log_reasoning(
            agent_id=self.agent_id,
            task_description=step_type,
            thinking_steps=[reasoning],
            final_conclusion=str(data),
            model="gemini-pro",
            metadata={'confidence': confidence}
            )
        
            # Send to UI callback
        if self.update_callback:
            self.update_callback(update_data)
    
    def get_linkedin_profile_data(self, username: str) -> Dict[str, Any]:
    """Get LinkedIn profile data with transparent reasoning"""
        
        self._send_reasoning_update(
        "profile_extraction_start",
        f"I need to extract comprehensive profile data for {username} to understand their professional context, career trajectory, and current situation.",
        "Starting LinkedIn profile data extraction process."
        )
        
        try:
        self._send_reasoning_update(
            "api_call_linkedin",
            f"Calling LinkedIn API to get profile data for {username}. This will provide experience, skills, education, and current role information that I need for wellness analysis.",
            "Making API call to LinkedIn profile endpoint."
        )
        
        response = self.linkedin_client.call_api('LinkedIn/get_user_profile_by_username', 
        query={'username': username})
        
        if response and (response.get('success') or response.get('id')):
            # Handle both wrapped and direct response formats
            profile_data = response.get('data', response) if response.get('success') else response
                
            self._send_reasoning_update(
            "data_extraction_success",
            f"Successfully extracted profile data for {username}. Found {len(profile_data.get('position', []))} work experiences, {len(profile_data.get('skills', []))} skills, and {len(profile_data.get('educations', []))} education entries.",
            "Profile data extraction completed successfully.",
            data={'profile_summary': self._create_profile_summary(profile_data)}
            )
                
        return profile_data
        else:
        self._send_reasoning_update(
            "data_extraction_error",
            f"LinkedIn API returned an error or empty response for {username}. This could mean the profile is private, doesn't exist, or there's an API issue.",
            "Will proceed with demo mode using sample data.",
        data={'error': response}
            )
        return self._get_demo_profile_data(username)
                
        except Exception as e:
        self._send_reasoning_update(
            "data_extraction_exception",
            f"Exception occurred while extracting profile data for {username}: {str(e)}. This could be due to network issues, API limits, or invalid username.",
            "Will proceed with demo mode to demonstrate the reasoning process.",
        data={'error': str(e)}
        )
        return self._get_demo_profile_data(username)
    
    def _create_profile_summary(self, profile_data: Dict) -> Dict:
    """Create a summary of key profile information"""
        return {
        'name': f"{profile_data.get('firstName', '')} {profile_data.get('lastName', '')}".strip(),
        'headline': profile_data.get('headline', 'N/A'),
        'location': profile_data.get('geo', {}).get('full', 'N/A'),
        'current_position': profile_data.get('position', [{}])[0].get('title', 'N/A') if profile_data.get('position') else 'N/A',
        'current_company': profile_data.get('position', [{}])[0].get('companyName', 'N/A') if profile_data.get('position') else 'N/A',
        'experience_count': len(profile_data.get('position', [])),
        'skills_count': len(profile_data.get('skills', [])),
        'education_count': len(profile_data.get('educations', []))
        }
    
    def _get_demo_profile_data(self, username: str) -> Dict[str, Any]:
    """Generate demo profile data for demonstration purposes"""
        return {
        'firstName': 'Demo',
        'lastName': 'User',
        'username': username,
        'headline': 'Senior Software Engineer | AI/ML Enthusiast | Tech Leadership',
        'geo': {'full': 'San Francisco, CA'},
        'position': [
            {
            'title': 'Senior Software Engineer',
            'companyName': 'TechCorp Inc.',
            'start': {'year': 2022},
            'end': {'year': 0},  # Current position
            'description': 'Leading AI/ML initiatives and managing a team of 5 engineers'
            },
            {
            'title': 'Software Engineer',
            'companyName': 'StartupXYZ',
            'start': {'year': 2020},
            'end': {'year': 2022},
            'description': 'Full-stack development with focus on scalable backend systems'
            }
        ],
        'skills': [
            {'name': 'Python', 'endorsementsCount': 45},
            {'name': 'Machine Learning', 'endorsementsCount': 32},
            {'name': 'Leadership', 'endorsementsCount': 28},
            {'name': 'React', 'endorsementsCount': 25},
            {'name': 'AWS', 'endorsementsCount': 22}
        ],
        'educations': [
            {
            'schoolName': 'Stanford University',
            'degree': 'Master of Science',
            'fieldOfStudy': 'Computer Science',
            'start': {'year': 2018},
            'end': {'year': 2020}
            }
        ],
        'summary': 'Passionate software engineer with expertise in AI/ML and team leadership. Focused on building scalable solutions and mentoring junior developers.',
        'isDemo': True
        }
    
    def analyze_stress_factors_with_reasoning(self, profile_data: Dict) -> List[Dict]:
    """Analyze potential stress factors with transparent chain of thought reasoning"""
        
        if not BMASTERAI_AVAILABLE:
            return self._analyze_stress_factors_basic(profile_data)
        
            # cot = ChainOfThought(self.agent_id, "stress_factor_identification", "gemini-pro")
            stress_factors = []
        
            # Career Transition Analysis
            # cot.step("Analyzing career transitions and job stability")
            # cot.reason("Recent job changes can indicate career instability, growth seeking, or adaptation to market changes. Each scenario has different stress implications.")
        
            positions = profile_data.get('position', [])
        if len(positions) > 1:
            recent_changes = sum(1 for pos in positions[:3] if pos.get('end', {}).get('year', 0) > 2020)
        if recent_changes >= 2:
            stress_factor = {
            'category': 'Career Instability',
            'severity': 'Medium',
            'description': f'Multiple job changes in recent years ({recent_changes} changes since 2020)',
            'reasoning': 'Frequent job changes can create uncertainty about career direction and financial stability, leading to increased stress levels.',
            'impact': 'May cause anxiety about job security and career progression'
            }
            stress_factors.append(stress_factor)
                
            self._send_reasoning_update(
            "career_transition_analysis",
            f"Identified {recent_changes} job changes since 2020. This pattern suggests either rapid career growth or potential instability.",
            f"Classified as medium-severity stress factor due to potential uncertainty.",
            data=stress_factor
            )
        
            # Industry and Role Pressure Assessment
            # cot.step("Evaluating industry-specific and role-specific stress factors")
            # cot.reason("Different industries and roles have varying stress levels, work cultures, and pressure points that affect well-being.")
        
            current_title = positions[0].get('title', '') if positions else ''
            current_company = positions[0].get('companyName', '') if positions else ''
        
            # Check for high-stress role indicators
            high_stress_keywords = ['senior', 'lead', 'manager', 'director', 'vp', 'cto', 'ceo']
        if any(keyword in current_title.lower() for keyword in high_stress_keywords):
            stress_factor = {
            'category': 'Leadership Responsibility',
            'severity': 'High',
            'description': f'Senior role with leadership responsibilities: {current_title}',
            'reasoning': 'Leadership positions typically involve higher stress due to increased responsibility, decision-making pressure, and accountability for team performance.',
            'impact': 'May experience pressure from managing people, meeting targets, and making critical decisions'
            }
            stress_factors.append(stress_factor)
        
            self._send_reasoning_update(
            "leadership_stress_analysis",
            f"Current role '{current_title}' indicates leadership responsibilities. Senior positions typically involve higher stress levels due to increased accountability.",
            "Identified high-severity stress factor related to leadership pressure.",
            data=stress_factor
            )
        
            # Tech Industry Specific Stress
            tech_indicators = ['engineer', 'developer', 'software', 'tech', 'ai', 'ml', 'data']
        if any(indicator in current_title.lower() for indicator in tech_indicators):
            stress_factor = {
            'category': 'Tech Industry Pressure',
            'severity': 'Medium',
            'description': 'Working in fast-paced technology sector',
            'reasoning': 'Tech industry is known for rapid changes, continuous learning requirements, long hours, and high performance expectations.',
            'impact': 'May experience pressure to stay current with technology, meet tight deadlines, and handle complex problem-solving'
            }
            stress_factors.append(stress_factor)
        
            self._send_reasoning_update(
            "tech_industry_analysis",
            "Identified technology sector employment. Tech industry typically involves rapid change, continuous learning pressure, and high-performance expectations.",
            "Added medium-severity stress factor for tech industry pressures.",
            data=stress_factor
            )
        
            # Work-Life Balance Assessment
            # cot.step("Examining work-life balance indicators")
            # cot.reason("Location, company size, and role demands significantly affect personal time and work-life balance.")
        
            location = profile_data.get('geo', {}).get('full', '')
            high_cost_areas = ['san francisco', 'new york', 'seattle', 'boston', 'los angeles']
        if any(area in location.lower() for area in high_cost_areas):
            stress_factor = {
            'category': 'Cost of Living Pressure',
            'severity': 'Medium',
            'description': f'Living in high-cost area: {location}',
            'reasoning': 'High cost of living areas create financial pressure and may require longer work hours or higher salaries to maintain lifestyle.',
            'impact': 'Financial stress and potential for longer commutes or smaller living spaces'
            }
            stress_factors.append(stress_factor)
        
            self._send_reasoning_update(
            "location_stress_analysis",
            f"Location '{location}' is identified as a high-cost living area. This typically creates additional financial pressure and lifestyle constraints.",
            "Added medium-severity stress factor for cost of living pressures.",
            data=stress_factor
            )
        
            # Skill Gap Analysis
            # cot.step("Assessing potential skill gaps and learning pressure")
            # cot.reason("In rapidly evolving fields, skill gaps can create anxiety about career relevance and job security.")
        
            skills = profile_data.get('skills', [])
        if len(skills) < 10:
            stress_factor = {
            'category': 'Skill Development Pressure',
            'severity': 'Low',
            'description': f'Limited skill diversity ({len(skills)} skills listed)',
            'reasoning': 'Having fewer documented skills may indicate need for skill development or better self-promotion, both of which can create professional anxiety.',
            'impact': 'May feel pressure to develop more skills or better showcase existing capabilities'
            }
            stress_factors.append(stress_factor)
        
            self._send_reasoning_update(
            "skill_gap_analysis",
            f"Profile shows {len(skills)} documented skills. This may indicate opportunity for skill development or better self-promotion.",
            "Added low-severity stress factor for skill development needs.",
            data=stress_factor
            )
        
            # cot.conclude(f"Identified {len(stress_factors)} potential stress factors across career, industry, location, and skill dimensions.")
        
        return stress_factors
    
    def _analyze_stress_factors_basic(self, profile_data: Dict) -> List[Dict]:
    """Basic stress factor analysis without BMasterAI"""
        stress_factors = []
        
        # Simple analysis without chain of thought
        positions = profile_data.get('position', [])
        current_title = positions[0].get('title', '') if positions else ''
        
        if 'senior' in current_title.lower() or 'lead' in current_title.lower():
            stress_factors.append({
            'category': 'Leadership Responsibility',
            'severity': 'High',
            'description': f'Senior role: {current_title}',
            'reasoning': 'Leadership positions involve higher stress and responsibility.',
            'impact': 'Increased pressure and decision-making responsibility'
            })
        
        return stress_factors
    
    def identify_happiness_opportunities_with_reasoning(self, profile_data: Dict, stress_factors: List[Dict]) -> List[Dict]:
    """Identify happiness enhancement opportunities with transparent reasoning"""
        
        if not BMASTERAI_AVAILABLE:
            return self._identify_happiness_opportunities_basic(profile_data, stress_factors)
        
            # cot = ChainOfThought(self.agent_id, "happiness_opportunity_identification", "gemini-pro")
            opportunities = []
        
            # Skill Development and Confidence Building
            # cot.step("Identifying skill development opportunities for confidence building")
            # cot.reason("Continuous learning and skill building enhance career satisfaction, reduce anxiety about job security, and provide sense of accomplishment.")
        
            skills = profile_data.get('skills', [])
            current_title = profile_data.get('position', [{}])[0].get('title', '') if profile_data.get('position') else ''
        
            # Identify trending skills in their field
        if 'engineer' in current_title.lower() or 'developer' in current_title.lower():
            opportunity = {
            'category': 'Professional Development',
            'priority': 'High',
            'title': 'Expand Technical Leadership Skills',
            'description': 'Develop advanced technical leadership and mentoring capabilities',
            'reasoning': 'Technical professionals often find fulfillment in mentoring others and leading technical initiatives, which also enhances career prospects.',
            'specific_actions': [
            'Consider technical mentoring or coaching opportunities',
            'Explore speaking at tech meetups or conferences',
            'Lead a technical initiative or open-source project',
            'Develop skills in emerging technologies relevant to your field'
            ],
            'happiness_impact': 'Provides sense of purpose, recognition, and career advancement'
            }
            opportunities.append(opportunity)
        
            self._send_reasoning_update(
            "skill_development_opportunity",
            f"Based on technical role '{current_title}', identified opportunity for technical leadership development. This aligns with natural career progression and provides fulfillment through mentoring.",
            "Added high-priority opportunity for technical leadership skills.",
            data=opportunity
            )
        
            # Work-Life Balance Enhancement
            # cot.step("Assessing work-life balance improvement opportunities")
            # cot.reason("Better boundaries and time management lead to reduced stress and increased personal fulfillment without sacrificing career growth.")
        
            # Check if they have stress factors that could be mitigated
            has_high_stress = any(sf['severity'] == 'High' for sf in stress_factors)
        if has_high_stress:
            opportunity = {
            'category': 'Work-Life Balance',
            'priority': 'High',
            'title': 'Implement Stress Management and Boundary Setting',
            'description': 'Develop sustainable work practices and stress management techniques',
            'reasoning': 'Given identified high-stress factors, implementing structured stress management and boundary setting is crucial for long-term happiness and career sustainability.',
            'specific_actions': [
            'Establish clear work-life boundaries (e.g., no emails after 7 PM)',
            'Practice time-blocking for focused work and personal time',
            'Implement regular stress-relief activities (exercise, meditation, hobbies)',
            'Consider delegating more responsibilities if in leadership role'
            ],
            'happiness_impact': 'Reduces burnout risk, improves personal relationships, and increases overall life satisfaction'
            }
            opportunities.append(opportunity)
        
            self._send_reasoning_update(
            "work_life_balance_opportunity",
            "Identified high-stress factors in profile analysis. Implementing structured work-life balance improvements could significantly reduce stress and increase happiness.",
            "Added high-priority opportunity for stress management and boundary setting.",
            data=opportunity
            )
        
            # Career Alignment and Purpose
            # cot.step("Evaluating career-passion alignment and purpose fulfillment")
            # cot.reason("When work aligns with personal interests and values, job satisfaction and happiness increase significantly, leading to better performance and life satisfaction.")
        
            education = profile_data.get('educations', [])
        if education:
            latest_education = education[0]
            field_of_study = latest_education.get('fieldOfStudy', '')
        
            opportunity = {
            'category': 'Career Fulfillment',
            'priority': 'Medium',
            'title': 'Align Work with Educational Background and Interests',
            'description': f'Leverage {field_of_study} background for more fulfilling work',
            'reasoning': f'Educational background in {field_of_study} suggests deep interest in this area. Aligning current work more closely with this foundation can increase job satisfaction.',
            'specific_actions': [
            f'Seek projects that utilize {field_of_study} knowledge more directly',
            'Consider specializing in areas that combine current role with educational background',
            'Explore opportunities to teach or share knowledge in your field of study',
            'Join professional associations related to your educational background'
            ],
            'happiness_impact': 'Increases sense of purpose and utilizes natural interests and abilities'
            }
            opportunities.append(opportunity)
        
            self._send_reasoning_update(
            "career_alignment_opportunity",
            f"Educational background in {field_of_study} provides foundation for career alignment. Leveraging this background more directly could increase job satisfaction and sense of purpose.",
            "Added medium-priority opportunity for career-education alignment.",
            data=opportunity
            )
        
            # Social Connection and Networking
            # cot.step("Identifying social connection and professional networking opportunities")
            # cot.reason("Strong professional and personal relationships are key predictors of happiness and career success, providing support, opportunities, and sense of belonging.")
        
            location = profile_data.get('geo', {}).get('full', '')
        if location:
            opportunity = {
            'category': 'Social Connection',
            'priority': 'Medium',
            'title': 'Build Professional and Personal Community',
            'description': f'Expand professional network and social connections in {location}',
            'reasoning': 'Strong social connections are fundamental to happiness and career success. Building community provides support, opportunities, and sense of belonging.',
            'specific_actions': [
            f'Join professional meetups and organizations in {location}',
            'Participate in industry conferences and networking events',
            'Consider joining hobby groups or volunteer organizations',
            'Schedule regular coffee meetings with colleagues and industry peers'
            ],
            'happiness_impact': 'Provides social support, reduces isolation, and creates opportunities for collaboration and friendship'
            }
            opportunities.append(opportunity)
        
            self._send_reasoning_update(
            "social_connection_opportunity",
            f"Location in {location} provides opportunities for professional and social networking. Building stronger community connections can significantly impact happiness and career growth.",
            "Added medium-priority opportunity for community building.",
            data=opportunity
            )
        
            # Recognition and Achievement
            # cot.step("Identifying opportunities for recognition and achievement satisfaction")
            # cot.reason("Recognition and achievement are important motivators. Creating opportunities for visible accomplishments can boost confidence and career satisfaction.")
        
            skills = profile_data.get('skills', [])
            top_skills = sorted(skills, key=lambda x: x.get('endorsementsCount', 0), reverse=True)[:3]
        
        if top_skills:
            top_skill = top_skills[0]['name']
            opportunity = {
            'category': 'Recognition & Achievement',
            'priority': 'Medium',
            'title': f'Showcase Expertise in {top_skill}',
            'description': f'Build recognition around your strongest skill: {top_skill}',
            'reasoning': f'{top_skill} appears to be a key strength with {top_skills[0].get("endorsementsCount", 0)} endorsements. Building visible expertise in this area can provide recognition and career advancement.',
            'specific_actions': [
            f'Write articles or blog posts about {top_skill}',
            f'Speak at conferences or meetups about {top_skill}',
            f'Mentor others in {top_skill}',
            f'Contribute to open-source projects showcasing {top_skill}'
            ],
            'happiness_impact': 'Provides recognition, builds reputation, and creates sense of expertise and authority'
            }
            opportunities.append(opportunity)
        
            self._send_reasoning_update(
            "recognition_opportunity",
            f"Top skill '{top_skill}' with {top_skills[0].get('endorsementsCount', 0)} endorsements represents a strength that could be leveraged for greater recognition and career satisfaction.",
            "Added medium-priority opportunity for expertise showcase.",
            data=opportunity
            )
        
            # cot.conclude(f"Identified {len(opportunities)} happiness enhancement opportunities across professional development, work-life balance, career alignment, social connection, and recognition dimensions.")
        
        return opportunities
    
    def _identify_happiness_opportunities_basic(self, profile_data: Dict, stress_factors: List[Dict]) -> List[Dict]:
    """Basic happiness opportunity identification without BMasterAI"""
        opportunities = []
        
        # Simple analysis
        skills = profile_data.get('skills', [])
        if skills:
            top_skill = max(skills, key=lambda x: x.get('endorsementsCount', 0))['name']
            opportunities.append({
            'category': 'Professional Development',
            'priority': 'High',
            'title': f'Develop expertise in {top_skill}',
            'description': f'Build on your strength in {top_skill}',
            'reasoning': 'Leveraging existing strengths builds confidence.',
            'specific_actions': [f'Take advanced courses in {top_skill}'],
            'happiness_impact': 'Increased confidence and career prospects'
            })
        
        return opportunities
    
    def generate_personalized_recommendations_with_reasoning(self, profile_data: Dict, stress_factors: List[Dict], opportunities: List[Dict]) -> List[Dict]:
    """Generate personalized recommendations with transparent reasoning"""
        
        if not BMASTERAI_AVAILABLE:
            return self._generate_recommendations_basic(profile_data, stress_factors, opportunities)
        
            # cot = ChainOfThought(self.agent_id, "personalized_recommendation_generation", "gemini-pro")
            recommendations = []
        
            # Prioritization and Integration
            # cot.step("Prioritizing recommendations based on impact and feasibility")
            # cot.reason("I need to prioritize recommendations based on potential impact, feasibility, and alignment with identified stress factors and happiness opportunities.")
        
            # High-priority recommendations addressing immediate stress
            high_stress_factors = [sf for sf in stress_factors if sf['severity'] == 'High']
            high_priority_opportunities = [op for op in opportunities if op['priority'] == 'High']
        
        if high_stress_factors and high_priority_opportunities:
            # Create integrated recommendation addressing both stress and opportunity
            stress_factor = high_stress_factors[0]
            opportunity = high_priority_opportunities[0]
        
            recommendation = {
            'title': 'Immediate Stress Management with Skill Development',
            'priority': 'Critical',
            'timeframe': '1-3 months',
            'description': f'Address {stress_factor["category"].lower()} while building {opportunity["category"].lower()} capabilities',
            'reasoning': f'Combining stress management for {stress_factor["category"]} with {opportunity["category"]} development creates a sustainable approach that addresses immediate concerns while building long-term satisfaction.',
            'specific_steps': [
            {
            'step': 1,
            'action': 'Assess current stress levels and triggers',
            'timeline': 'Week 1',
            'reasoning': 'Understanding specific stress triggers enables targeted interventions'
            },
            {
            'step': 2,
            'action': 'Implement daily stress management routine',
            'timeline': 'Week 2-4',
            'reasoning': 'Consistent daily practices build resilience and coping capacity'
            },
            {
            'step': 3,
            'action': f'Begin {opportunity["title"].lower()} activities',
            'timeline': 'Week 3-8',
            'reasoning': 'Starting skill development early provides positive focus and future benefits'
            },
            {
            'step': 4,
            'action': 'Evaluate progress and adjust approach',
            'timeline': 'Week 8-12',
            'reasoning': 'Regular evaluation ensures the approach remains effective and sustainable'
            }
            ],
            'expected_outcomes': [
            'Reduced daily stress levels',
            'Improved coping mechanisms',
            'Enhanced professional capabilities',
            'Greater sense of control and direction'
            ],
            'success_metrics': [
            'Daily stress rating (1-10 scale)',
            'Sleep quality improvement',
            'Progress on skill development goals',
            'Feedback from colleagues or supervisors'
            ]
            }
            recommendations.append(recommendation)
        
            self._send_reasoning_update(
            "critical_recommendation_generation",
            f"Created integrated recommendation addressing high-severity stress factor ({stress_factor['category']}) with high-priority opportunity ({opportunity['category']}). This approach tackles immediate concerns while building long-term satisfaction.",
            "Generated critical-priority recommendation combining stress management with skill development.",
            data=recommendation
            )
        
            # Medium-term career development recommendation
            # cot.step("Creating medium-term career development recommendations")
            # cot.reason("Medium-term recommendations should focus on sustainable career growth and life satisfaction improvements that build on immediate stress management.")
        
            current_title = profile_data.get('position', [{}])[0].get('title', '') if profile_data.get('position') else ''
            skills = profile_data.get('skills', [])
        
            recommendation = {
            'title': 'Strategic Career Development and Life Integration',
            'priority': 'High',
            'timeframe': '3-12 months',
            'description': 'Build sustainable career growth while maintaining work-life balance',
            'reasoning': f'Based on current role as {current_title} and skill set, focusing on strategic career development with life integration will provide long-term satisfaction and reduced stress.',
            'specific_steps': [
            {
            'step': 1,
            'action': 'Define 12-month career and life goals',
            'timeline': 'Month 1',
            'reasoning': 'Clear goals provide direction and motivation while preventing drift'
            },
            {
            'step': 2,
            'action': 'Identify skill gaps and development opportunities',
            'timeline': 'Month 1-2',
            'reasoning': 'Understanding skill gaps enables targeted learning and career advancement'
            },
            {
            'step': 3,
            'action': 'Create learning and networking plan',
            'timeline': 'Month 2-3',
            'reasoning': 'Structured learning and networking accelerate career growth and satisfaction'
            },
            {
            'step': 4,
            'action': 'Implement monthly progress reviews',
            'timeline': 'Month 3-12',
            'reasoning': 'Regular reviews ensure progress and allow for course corrections'
            }
            ],
            'expected_outcomes': [
            'Clear career direction and goals',
            'Enhanced professional network',
            'Improved skills and capabilities',
            'Better work-life integration'
            ],
            'success_metrics': [
            'Achievement of monthly learning goals',
            'Expansion of professional network',
            'Positive feedback on new skills',
            'Improved work-life balance rating'
            ]
            }
            recommendations.append(recommendation)
        
            self._send_reasoning_update(
            "career_development_recommendation",
            f"Created strategic career development recommendation based on current role '{current_title}' and identified opportunities. This provides sustainable growth path while maintaining life balance.",
            "Generated high-priority recommendation for strategic career development.",
            data=recommendation
            )
        
            # Long-term fulfillment and purpose recommendation
            # cot.step("Developing long-term fulfillment and purpose recommendations")
            # cot.reason("Long-term recommendations should focus on deeper fulfillment, purpose, and sustainable happiness that goes beyond immediate career concerns.")
        
            education = profile_data.get('educations', [])
            location = profile_data.get('geo', {}).get('full', '')
        
            recommendation = {
            'title': 'Purpose-Driven Impact and Community Building',
            'priority': 'Medium',
            'timeframe': '6-24 months',
            'description': 'Build meaningful impact and community connections for long-term fulfillment',
            'reasoning': 'Long-term happiness comes from sense of purpose, meaningful relationships, and positive impact. Building these elements creates sustainable fulfillment beyond career success.',
            'specific_steps': [
            {
            'step': 1,
            'action': 'Identify personal values and impact areas',
            'timeline': 'Month 1-2',
            'reasoning': 'Understanding personal values guides meaningful choices and activities'
            },
            {
            'step': 2,
            'action': 'Find mentoring or teaching opportunities',
            'timeline': 'Month 3-6',
            'reasoning': 'Sharing knowledge and helping others provides deep satisfaction and purpose'
            },
            {
            'step': 3,
            'action': 'Join or create community initiatives',
            'timeline': 'Month 6-12',
            'reasoning': 'Community involvement builds relationships and creates positive impact'
            },
            {
            'step': 4,
            'action': 'Develop long-term impact projects',
            'timeline': 'Month 12-24',
            'reasoning': 'Long-term projects provide sustained purpose and meaningful achievement'
            }
            ],
            'expected_outcomes': [
            'Stronger sense of purpose and meaning',
            'Expanded community connections',
            'Positive impact on others',
            'Enhanced life satisfaction and fulfillment'
            ],
            'success_metrics': [
            'Number of people mentored or helped',
            'Community involvement level',
            'Personal fulfillment rating',
            'Long-term project progress'
            ]
            }
            recommendations.append(recommendation)
        
            self._send_reasoning_update(
            "purpose_fulfillment_recommendation",
            f"Created long-term fulfillment recommendation focusing on purpose, community, and impact. This addresses deeper happiness needs beyond immediate career concerns.",
            "Generated medium-priority recommendation for purpose-driven impact.",
            data=recommendation
            )
        
            # cot.conclude(f"Generated {len(recommendations)} personalized recommendations spanning immediate stress management, strategic career development, and long-term purpose fulfillment.")
        
        return recommendations
    
    def _generate_recommendations_basic(self, profile_data: Dict, stress_factors: List[Dict], opportunities: List[Dict]) -> List[Dict]:
    """Basic recommendation generation without BMasterAI"""
        recommendations = []
        
        # Simple recommendation based on stress factors
        if stress_factors:
            recommendations.append({
            'title': 'Stress Management Plan',
            'priority': 'High',
            'timeframe': '1-3 months',
            'description': 'Address identified stress factors',
            'reasoning': 'Managing stress is crucial for well-being.',
            'specific_steps': [
            {'step': 1, 'action': 'Identify stress triggers', 'timeline': 'Week 1'},
            {'step': 2, 'action': 'Implement coping strategies', 'timeline': 'Week 2-12'}
            ],
            'expected_outcomes': ['Reduced stress levels'],
            'success_metrics': ['Daily stress rating']
            })
        
        return recommendations
    
    def analyze_linkedin_profile_for_wellness(self, username: str) -> Dict[str, Any]:
    """
        Complete LinkedIn profile analysis for wellness recommendations with full chain of thought transparency
    """
        
        if not BMASTERAI_AVAILABLE:
            return self._analyze_profile_basic(username)
        
            with ReasoningSession(
            self.agent_id,
            f"Analyze LinkedIn profile {username} for personalized stress reduction and happiness suggestions",
            self.model_name,
            metadata={"username": username, "analysis_type": "wellness_assessment"}
            ) as session:
        
            # Step 1: Profile Data Extraction
            session.think(
            f"I need to extract comprehensive profile data for {username} to understand their professional context, career trajectory, and potential stress factors. This will form the foundation for personalized wellness recommendations.",
            confidence=0.9
            )
        
            profile_data = self.get_linkedin_profile_data(username)
        
        if not profile_data:
            session.think("Profile data extraction failed. I'll provide general wellness guidance and demonstrate the reasoning process.", confidence=0.5)
        return {'error': 'Profile data not available', 'demo_mode': True}
        
        # Step 2: Stress Factor Identification
        session.think(
            "Now I'll analyze the profile data to identify potential sources of stress based on role, industry, career changes, and work-life balance indicators. This systematic analysis will help me understand what might be causing stress in their professional life.",
        confidence=0.8
        )
        
        stress_factors = self.analyze_stress_factors_with_reasoning(profile_data)
        
        # Step 3: Happiness Opportunity Assessment
        session.think(
            "I'll look for opportunities to enhance happiness through career satisfaction, skill development, work-life balance, and personal fulfillment. These opportunities should complement the stress reduction strategies.",
        confidence=0.8
        )
        
        happiness_opportunities = self.identify_happiness_opportunities_with_reasoning(profile_data, stress_factors)
        
        # Step 4: Personalized Recommendation Generation
        session.think(
            "Based on my analysis of stress factors and happiness opportunities, I'll generate specific, actionable recommendations tailored to this person's unique professional situation. These recommendations will be prioritized and include concrete steps.",
        confidence=0.9
        )
        
        recommendations = self.generate_personalized_recommendations_with_reasoning(
            profile_data, stress_factors, happiness_opportunities
        )
        
        # Step 5: Final Integration and Summary
        session.think(
            f"I've completed a comprehensive analysis identifying {len(stress_factors)} stress factors, {len(happiness_opportunities)} happiness opportunities, and generated {len(recommendations)} personalized recommendations. Now I'll create a cohesive summary with clear next steps.",
        confidence=0.9
        )
        
        # Create comprehensive analysis result
        analysis_result = {
            'profile_summary': self._create_profile_summary(profile_data),
            'stress_factors': stress_factors,
            'happiness_opportunities': happiness_opportunities,
            'recommendations': recommendations,
            'analysis_metadata': {
            'username': username,
            'analysis_date': datetime.now().isoformat(),
            'total_stress_factors': len(stress_factors),
            'total_opportunities': len(happiness_opportunities),
            'total_recommendations': len(recommendations),
            'high_priority_recommendations': len([r for r in recommendations if r['priority'] in ['Critical', 'High']]),
            'is_demo': profile_data.get('isDemo', False)
            }
        }
        
        session.decide(
            "Analysis completion assessment",
            ["comprehensive_analysis_complete", "partial_analysis_complete", "analysis_insufficient"],
            "comprehensive_analysis_complete",
            f"Successfully completed comprehensive wellness analysis with {len(stress_factors)} stress factors, {len(happiness_opportunities)} opportunities, and {len(recommendations)} actionable recommendations."
        )
        
        self._send_reasoning_update(
            "analysis_complete",
            f"Completed comprehensive LinkedIn wellness analysis for {username}. Generated personalized recommendations addressing both stress reduction and happiness enhancement.",
            f"Analysis complete with {len(recommendations)} actionable recommendations.",
        data=analysis_result['analysis_metadata']
        )
        
        return analysis_result
    
    def _analyze_profile_basic(self, username: str) -> Dict[str, Any]:
    """Basic profile analysis without BMasterAI"""
        profile_data = self.get_linkedin_profile_data(username)
        stress_factors = self._analyze_stress_factors_basic(profile_data)
        opportunities = self._identify_happiness_opportunities_basic(profile_data, stress_factors)
        recommendations = self._generate_recommendations_basic(profile_data, stress_factors, opportunities)
        
        return {
        'profile_summary': self._create_profile_summary(profile_data),
        'stress_factors': stress_factors,
        'happiness_opportunities': opportunities,
        'recommendations': recommendations,
        'analysis_metadata': {
            'username': username,
            'analysis_date': datetime.now().isoformat(),
            'total_stress_factors': len(stress_factors),
            'total_opportunities': len(opportunities),
            'total_recommendations': len(recommendations),
            'is_demo': profile_data.get('isDemo', False)
        }
        }
    
    def get_agent_stats(self) -> Dict[str, Any]:
    """Get agent statistics for analytics display"""
        if BMASTERAI_AVAILABLE and self.monitor:
            return {
            'total_events': getattr(self.monitor, 'total_events', 0),
            'event_types': getattr(self.monitor, 'event_types', {}),
            'bmasterai_available': True
            }
        else:
        return {
            'total_events': 0,
            'event_types': {},
            'bmasterai_available': False
        }
    
    def export_reasoning_logs(self, format_type: str = "json") -> str:
    """Export reasoning logs in specified format"""
        if not BMASTERAI_AVAILABLE:
            return "BMasterAI not available - no logs to export"
        
            # This would integrate with BMasterAI's export functionality
            # For now, return a placeholder
        if format_type == "json":
            return json.dumps({
            'agent_id': self.agent_id,
            'export_date': datetime.now().isoformat(),
            'note': 'BMasterAI log export functionality would be integrated here'
            }, indent=2)
        else:
        return f"# BMasterAI Reasoning Logs\n\nAgent ID: {self.agent_id}\nExport Date: {datetime.now().isoformat()}\n\nBMasterAI log export functionality would be integrated here."




# -*- coding: utf-8 -*-

# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""RAG retrieval tool for content generation."""

import re
from typing import List, Optional, Union
import warnings

from google.cloud import aiplatform_v1beta1
from google.cloud.aiplatform import initializer
from google.cloud.aiplatform_v1beta1.types import tool as gapic_tool_types
from vertexai.preview import generative_models
from vertexai.preview.rag.utils import _gapic_utils
from vertexai.preview.rag.utils import resources


class Retrieval(generative_models.grounding.Retrieval):
    """Defines a retrieval tool that a model can call to access external knowledge."""

    def __init__(
        self,
        source: Union["VertexRagStore"],
        disable_attribution: Optional[bool] = False,
    ):
        self._raw_retrieval = gapic_tool_types.Retrieval(
            vertex_rag_store=source._raw_vertex_rag_store,
            disable_attribution=disable_attribution,
        )


class VertexRagStore:
    """Retrieve from Vertex RAG Store."""

    def __init__(
        self,
        rag_resources: Optional[List[resources.RagResource]] = None,
        rag_corpora: Optional[List[str]] = None,
        similarity_top_k: Optional[int] = None,
        vector_distance_threshold: Optional[float] = None,
        rag_retrieval_config: Optional[resources.RagRetrievalConfig] = None,
    ):
        """Initializes a Vertex RAG store tool.

        Example usage:
        ```
        import vertexai

        vertexai.init(project="my-project")

        # Using deprecated parameters
        tool = Tool.from_retrieval(
            retrieval=vertexai.preview.rag.Retrieval(
                source=vertexai.preview.rag.VertexRagStore(
                    rag_corpora=["projects/my-project/locations/us-central1/ragCorpora/rag-corpus-1"],
                    similarity_top_k=3,
                    vector_distance_threshold=0.4,
                ),
            )
        )

        # Using RagRetrievalConfig. Equivalent to the above example.
        config = vertexai.preview.rag.RagRetrievalConfig(
            top_k=2,
            filter=vertexai.preview.rag.RagRetrievalConfig.Filter(
                vector_distance_threshold=0.5
            ),
        )

        tool = Tool.from_retrieval(
            retrieval=vertexai.preview.rag.Retrieval(
                source=vertexai.preview.rag.VertexRagStore(
                    rag_corpora=["projects/my-project/locations/us-central1/ragCorpora/rag-corpus-1"],
                    rag_retrieval_config=config,
                ),
            )
        )
        ```

        Args:
            rag_resources: List of RagResource to retrieve from. It can be used
                to specify corpus only or ragfiles. Currently only support one
                corpus or multiple files from one corpus. In the future we
                may open up multiple corpora support.
            rag_corpora: If rag_resources is not specified, use rag_corpora as a
                list of rag corpora names. Deprecated. Use rag_resources instead.
            similarity_top_k: Number of top k results to return from the selected
                corpora. Deprecated. Use rag_retrieval_config.top_k instead.
            vector_distance_threshold (float):
                Optional. Only return results with vector distance smaller
                than the threshold. Deprecated. Use
                rag_retrieval_config.filter.vector_distance_threshold instead.
            rag_retrieval_config: Optional. The config containing the retrieval
                parameters, including top_k and vector_distance_threshold.
        """

        if rag_resources:
            if len(rag_resources) > 1:
                raise ValueError("Currently only support 1 RagResource.")
            name = rag_resources[0].rag_corpus
        elif rag_corpora:
            if len(rag_corpora) > 1:
                raise ValueError("Currently only support 1 RagCorpus.")
            warnings.warn(
                "rag_corpora is deprecated. Please use rag_resources instead."
                f" After {resources.DEPRECATION_DATE} using"
                " rag_corpora will raise error",
                DeprecationWarning,
            )
            name = rag_corpora[0]
        else:
            raise ValueError("rag_resources or rag_corpora must be specified.")

        data_client = _gapic_utils.create_rag_data_service_client()
        if data_client.parse_rag_corpus_path(name):
            rag_corpus_name = name
        elif re.match("^{}$".format(_gapic_utils._VALID_RESOURCE_NAME_REGEX), name):
            parent = initializer.global_config.common_location_path()
            rag_corpus_name = parent + "/ragCorpora/" + name
        else:
            raise ValueError(
                f"Invalid RagCorpus name: {rag_corpora}. Proper format should"
                + " be: projects/{{project}}/locations/{{location}}/ragCorpora/{{rag_corpus_id}}"
            )

        # Check for deprecated parameters and raise warnings.
        if similarity_top_k:
            # If similarity_top_k is specified, throw deprecation warning.
            warnings.warn(
                "similarity_top_k is deprecated. Please use"
                " rag_retrieval_config.top_k instead."
                f" After {resources.DEPRECATION_DATE} using"
                " similarity_top_k will raise error",
                DeprecationWarning,
            )
        if vector_distance_threshold:
            # If vector_distance_threshold is specified, throw deprecation warning.
            warnings.warn(
                "vector_distance_threshold is deprecated. Please use"
                " rag_retrieval_config.filter.vector_distance_threshold instead."
                f" After {resources.DEPRECATION_DATE} using"
                " vector_distance_threshold will raise error",
                DeprecationWarning,
            )

        # If rag_retrieval_config is not specified, set it to default values.
        if not rag_retrieval_config:
            api_retrival_config = aiplatform_v1beta1.RagRetrievalConfig(
                top_k=similarity_top_k,
                filter=aiplatform_v1beta1.RagRetrievalConfig.Filter(
                    vector_distance_threshold=vector_distance_threshold
                ),
            )
        else:
            # If rag_retrieval_config is specified, check for missing parameters.
            api_retrival_config = aiplatform_v1beta1.RagRetrievalConfig()
            # Set top_k to config value if specified
            if rag_retrieval_config.top_k:
                api_retrival_config.top_k = rag_retrieval_config.top_k
            else:
                api_retrival_config.top_k = similarity_top_k
            # Check if both vector_distance_threshold and vector_similarity_threshold
            # are specified.
            if (
                rag_retrieval_config.filter
                and rag_retrieval_config.filter.vector_distance_threshold
                and rag_retrieval_config.filter.vector_similarity_threshold
            ):
                raise ValueError(
                    "Only one of vector_distance_threshold or"
                    " vector_similarity_threshold can be specified at a time"
                    " in rag_retrieval_config."
                )
            # Set vector_distance_threshold to config value if specified
            if (
                rag_retrieval_config.filter
                and rag_retrieval_config.filter.vector_distance_threshold
            ):
                api_retrival_config.filter.vector_distance_threshold = (
                    rag_retrieval_config.filter.vector_distance_threshold
                )
            else:
                api_retrival_config.filter.vector_distance_threshold = (
                    vector_distance_threshold
                )
            # Set vector_similarity_threshold to config value if specified
            if (
                rag_retrieval_config.filter
                and rag_retrieval_config.filter.vector_similarity_threshold
            ):
                api_retrival_config.filter.vector_similarity_threshold = (
                    rag_retrieval_config.filter.vector_similarity_threshold
                )
            # Check if both rank_service and llm_ranker are specified.
            if (
                rag_retrieval_config.ranking
                and rag_retrieval_config.ranking.rank_service
                and rag_retrieval_config.ranking.rank_service.model_name
                and rag_retrieval_config.ranking.llm_ranker
                and rag_retrieval_config.ranking.llm_ranker.model_name
            ):
                raise ValueError(
                    "Only one of rank_service or llm_ranker can be specified"
                    " at a time in rag_retrieval_config."
                )
            # Set rank_service to config value if specified
            if (
                rag_retrieval_config.ranking
                and rag_retrieval_config.ranking.rank_service
            ):
                api_retrival_config.ranking.rank_service.model_name = (
                    rag_retrieval_config.ranking.rank_service.model_name
                )
            # Set llm_ranker to config value if specified
            if rag_retrieval_config.ranking and rag_retrieval_config.ranking.llm_ranker:
                api_retrival_config.ranking.llm_ranker.model_name = (
                    rag_retrieval_config.ranking.llm_ranker.model_name
                )

        if rag_resources:
            gapic_rag_resource = gapic_tool_types.VertexRagStore.RagResource(
                rag_corpus=rag_corpus_name,
                rag_file_ids=rag_resources[0].rag_file_ids,
            )
            self._raw_vertex_rag_store = gapic_tool_types.VertexRagStore(
                rag_resources=[gapic_rag_resource],
                rag_retrieval_config=api_retrival_config,
            )
        else:
            self._raw_vertex_rag_store = gapic_tool_types.VertexRagStore(
                rag_corpora=[rag_corpus_name],
                rag_retrieval_config=api_retrival_config,
            )
